#!/bin/sh
PATH=$JAVA_HOME/bin:$PATH


java -help >/dev/null 2>&1 
if test $? != "0" ; then
  if [ $JAVA_HOME = "" ] ; then
    echo "Error: Unable to find java."
    echo "       Either JAVA_HOME needs to be set or 'java' in your PATH"
    echo ""
  else
    PATH=$PATH:$JAVA_HOME/bin
    export PATH
  fi  
fi


totalpages=0
yearLevel=0   
commentlevel=1
genImage=y
fromPage=0
toPage=unlimited
javarun=y
ebookfile=
imageFormat="png"
firstSnapShot=


#font type
font3dtype=0
#font format: glf or bmf
font3dformat=glf
#font wire or solid
font3dform=solid


function print_usage {
  echo "Usage : ${0##*/} [-commentLevel num] [-genImage y/n] [-startPage num]"
  echo "               [-stopPage num] [-javarun y/n] [-tgaSnapshot filaname] ebookfile.[pdf|djvu]"
  exit 1
}
 
function parse_options {

  while : ; do
    case $# in 0 ) break ;; esac # no arguments
    case "$1" in
      -commentLevel )
        shift
        commentlevel=$1
        shift
       ;;
      -startPage )
        shift
        fromPage=$1
        shift
       ;;
      -stopPage )
        shift
        toPage=$1
        shift
       ;;
      -genImage )
        shift
        genImage=$1
        shift
       ;;
      -javarun )
        shift
        javarun=$1
        shift
       ;;
      -tgaSnapshot )
        shift
        firstSnapShot=$1
        shift
       ;;
      -- )     # Stop option processing
        shift
        break
       ;;
      -* )     # Stop option processing
        print_usage
        break
       ;;
      * )
        break
       ;;
    esac
  done

  if [ $# != "1" ] ; then
    print_usage
  fi
  ebookfile="$@"
}


PROGNAME=$0;
echo "$PROGNAME" 
A=${PROGNAME##/*}
echo "$A"



if [ ${PROGNAME##/*}x != "x" ] ; then
  PROGNAME=`pwd`/$PROGNAME
fi
echo "$PROGNAME" 


export OTB_HOME=${PROGNAME%/bin/script/*}
echo "OTB_HOME $OTB_HOME" 

OTB_JAR=$OTB_HOME/installed/otb.jar

GL4JAVA_HOME=$OTB_HOME/installed/GL4Java
ITEXT_HOME=$OTB_HOME/installed/itext
JMAGIC_HOME=$OTB_HOME/installed/JMagick
DJVU_HOME=$OTB_HOME/installed/djvulibre
#old
IMAGE_MAGICK_HOME=$OTB_HOME/installed/ImageMagick
#IMAGE_MAGICK_HOME=/research/ycc1/packages/ImageMaigck-5.5.7
GS_HOME=$OTB_HOME/installed/ghostscript

echo "GL4JAVA_HOME $GL4JAVA_HOME"
#old

export GS_LIB=$GS_HOME/share/ghostscript/fonts:$GS_HOME/share/ghostscript/8.14/lib
export MAGICK_HOME=$IMAGE_MAGICK_HOME

#export MAGICK_HOME=$IMAGE_MAGICK_HOME/lib/ImageMagick-5.5.7
#export GS_LIB=/research/ycc1/packages/ghostscript-8.00/share/ghostscript/8.00/lib
export LD_LIBRARY_PATH=$GL4JAVA_HOME/lib:$JMAGIC_HOME/lib:$DJVU_HOME/lib
export PATH=$DJVU_HOME/bin:$IMAGE_MAGICK_HOME/bin:$GS_HOME/bin:{$PATH}

parse_options "$@"

# for precompiled images
imageWidth=512
imageHeight=512
f=${ebookfile%.*} #remove filename extension
TYPE=${ebookfile##*.} # extract filename extension

if [ $commentlevel -gt 1 ] ; then
  echo "startPage  : $fromPage"
  echo "stopPage   : $toPage"
  echo "genImage   : $genImage"
  echo "Ebook file : $ebookfile"
  echo "Snap shot file : $firstSnapShot"  
fi


searchTotalpages()
{
    echo "searchTotalpages()"

    local ebookdir=$1
    local ebookfile=$2
    echo "ebookdir> $ebookdir"
    echo "ebookfile> $ebookfile"
    
    # if / it means full path name
    if [ "${ebookfile##/*}x" = "x" ] ; then
      full_ebookfile=$ebookfile
    else
      full_ebookfile="$ebookdir$ebookfile"
    fi
    
    echo "full_ebookfile> $full_ebookfile"
    echo "type> $TYPE"
    if [ $TYPE = "pdf" ] ; then
	totalpages=`pdfinfo "$full_ebookfile.pdf" | grep Pages | awk '{print $2}'`
    else
	totalpages=`djvused "$full_ebookfile.djvu" -e 'n'`
    fi

    echo "total pages> $totalpages"

    if [ $toPage = "unlimited" ] ; then
	toPage=$((totalpages-1))
    fi

    return 0
}

searchYear()
{
    local ebookdir=$1
    local ebookfile=$2

    if [ "${ebookfile##/*}x" = "x" ] ; then
      full_ebookfile=$ebookfile
    else
      full_ebookfile="$ebookdir$ebookfile"
    fi
    
    yearLevel=10
    return 0
}

# ebook source and page image directory
EBOOKDIR="$OTB_HOME/data/"
PAGEDIR="$OTB_HOME/pagecache/"


# binary files
GL4JAVA_BIN=$GL4JAVA_HOME/jre/lib/ext
ITEXT_BIN=$ITEXT_HOME/bin
JMAGIC_BIN=$JMAGIC_HOME/lib

searchTotalpages $EBOOKDIR $f 
if [ $commentlevel -gt 1 ] ; then
  echo Returned total Pages: $totalpages
fi

searchYear $EBOOKDIR $f 
if [ $commentlevel -gt 1 ] ; then
  echo Returned Year:$yearLevel
fi

if [ $genImage = "y" ] ; then
  $OTB_HOME/bin/script/ebook2image.sh -commentlevel $commentlevel \
      $TYPE $PAGEDIR $EBOOKDIR $fromPage $toPage $f \
      $imageFormat $imageWidth $imageHeight
  if [ $toPage -lt $(($totalpages-1)) ] ; then
    $OTB_HOME/bin/script/ebook2image.sh -commentlevel $commentlevel \
        $TYPE $PAGEDIR $EBOOKDIR $((totalpages-2)) $(($totalpages-1)) $f \
        $imageFormat $imageWidth $imageHeight
  fi
fi


JARS=$JMAGIC_BIN/jmagick.jar:$GL4JAVA_BIN/png.jar:$GL4JAVA_BIN/gl4java.jar:$GL4JAVA_BIN/gl4java-glffonts.jar:$GL4JAVA_BIN/gl4java-glutfonts.jar:$ITEXT_BIN/iText.jar:$OTB_JAR

#if [ "$firstSnapShot" = "" ]; then
if [ -n "$firstSnapShot" ]; then 
    ARGS_FRONT="-classpath $JARS TTPproject.Ttp -tgaSnapshot $firstSnapShot -prefix $OTB_HOME -datafile testdata.config"
else
    ARGS_FRONT="-classpath $JARS TTPproject.Ttp -prefix $OTB_HOME -datafile testdata.config"
fi
ARGS_BACK="-from $fromPage -to $toPage -totalPages $totalpages -commentlevel $commentlevel -yearlevel $yearLevel -window 1024 700 10 20 -viewport 1024 700 0 0"

#ARGS_FONTS="-3dfonts $font3dtype -3dfonts_format $font3dformat -3dfonts_form $font3dform"

ebookfile_tail=${ebookfile##*/}

# Parameters:
#  -prefix OTB_HOME_DIRECTORY 
#  -datafile configuration_file 
#  -ebookfile pages number EBOOK_FILE_NAME 
#  -from fromPage 
#  -to toPage
#  -totalPages totalPages number
#   printing off different levels of comments: -commentlevel
#     0: turn everything off 
#     1: show loading and initialization information
#     2: show other debug information & loading and initialization information


if [ $javarun = "y" ]; then
    if [ $TYPE = "pdf" ] ; then
	echo "run pdf"
	echo $ARGS_FRONT 
	echo $ebookfile_tail
	echo $ARGS_BACK
	java $ARGS_FRONT -pdffile 1 $ebookfile_tail $ARGS_BACK #$ARGS_FONTS
    else
	echo "run djvus"
	java $ARGS_FRONT -djvufile $ebookfile_tail $ARGS_BACK #$ARGS_FONTS
    fi
fi