# if this file is executed, /bin/sh is used, as we don't start with #!
# this should work under ash, bash, zsh, ksh, sh style shells.

# make sure we are sourced, and not run
package="OpenTheBook"

if test "$0" != "`echo $0 | sed s/setup\.bash//`" ; then
# if $0 contains "setup.bash" we've been run... $0 is shellname if sourced.
# One exception is zsh has an option to set it temporarily to the script name
  if test -z "$ZSH_NAME" ; then
  # we aren't using zsh
  script_not_sourced=true
  fi
fi

if test -n "$script_not_sourced" ; then
  echo "        Error: Make sure you source this script, not execute it. Eg:"
  echo "                $ source setup.bash"
  echo "        or"
  echo "                $ . ./setup.bash"
  echo "        not"
  echo "                $ ./setup.bash"
elif test -n "$OTB_HOME" ; then
  echo "Your environment is already set up for $package!"
elif test ! -f setup.bash ; then
  echo "You must source the script from within the $package home directory"
else

  javac -help >/dev/null 2>&1 
  if test $? != "0" ; then
    echo ""
    echo "Unable to find Java compiler 'javac' in PATH:"
    echo $PATH
    echo "You will not be able to compile Java code."
    echo ""
  fi



  OTB_HOME=`pwd`
  export OTB_HOME
  OTBOS=`uname -s | tr '[A-Z]' '[a-z]' | sed 's/^cygwin.*$/cygwin/'`
  export OTBOS

  PATH=$OTB_HOME/bin/script:$OTB_HOME/bin/$OTBOS:$PATH
  PATH=$OTB_HOME/installed/ghostscript/bin:$PATH

  export PATH

  JAVA_BIN=`which javac`
  JAVA_HOME=${JAVA_BIN%/*/javac}
  export JAVA_HOME


  GS_LIB=$OTB_HOME/packages/ghostscript/lib
  export GS_LIB

  if test -z "$LD_LIBRARY_PATH" ; then
    LD_LIBRARY_PATH=$OTB_HOME:$OTB_HOME/lib/:$OTB_HOME/installed/ImageMagick/lib
  else
    LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$OTB_HOME:$OTB_HOME/lib/:$OTB_HOME/installed/ImageMagick/lib
  fi
  export LD_LIBRARY_PATH

  echo "Your environment has successfully been set up to run $package"

fi
unset script_not_sourced
