package TTPproject;

import javax.vecmath.Point3d;

import gl4java.awt.GLCanvas;
import gl4java.GLFunc;
import gl4java.GLUFunc;
import gl4java.GLEnum;

/**
 * 
 * @author Jack(Yi-Chun), Chu
 * @version 0.0, June 2002
 */


public class AnimatedPage  extends Thread 
    implements GLEnum, CONSTANTS
{
    private Point3d [][] rightStartCtrl_points= null;
    private Point3d [][] rightFinishCtrl_points= null;
    private Point3d [][] rightIntermediateCtrl_points= null;

    private Point3d [][] leftStartCtrl_points= null;
    private Point3d [][] leftFinishCtrl_points= null;
    private Point3d [][] leftIntermediateCtrl_points= null;
    
    private int totalsteps= 200;// steps
    private long msecsUPdateInterval= 500;
    private boolean running= false;
    public boolean initializing= false;
    private boolean stop= false;

    private BezierPatch rightPage= null;
    private BezierPatch leftPage= null;

    private Configuration config= null;
    private BookConstructor bc= null;
    private int rightTextureName= NONE;
    private int leftTextureName= NONE;

    //changed
    private int rightMaterialName= NONE;
    private int leftMaterialName= NONE;
    
    private int left_page_num= NONE;
    private int right_page_num= NONE;
    
    private int direction= 0; // 0 means from 3d to reading plane , 1 means from reading plane to 3d

    //changed
    public AnimatedPage()
    {
	running= false;
    }


    public AnimatedPage(BookConstructor bc,
			Point3d [][] leftStartCtrl_pts,
			Point3d [][] rightStartCtrl_pts,
			Point3d [][] leftFinishCtrl_pts,
			Point3d [][] rightFinishCtrl_pts,
			int steps,
			double updateInterval,
			Configuration config, 
			int direction)
    {	
	if(Ttp.commentslevel >= 2){
	    System.out.println("Animated.java: AnimatedPage() ");
	}

	this.direction= direction;
	//control points
	this.config= config;
	this.rightFinishCtrl_points= new Point3d[4][4];
	this.rightStartCtrl_points= new Point3d[4][4];
	this.rightIntermediateCtrl_points= new Point3d[4][4];
	
	this.leftStartCtrl_points= new Point3d[4][4];
	this.leftFinishCtrl_points= new Point3d[4][4];
	this.leftIntermediateCtrl_points= new Point3d[4][4];
	
	this.bc= bc;
	this.totalsteps= steps;
	this.msecsUPdateInterval= (long) (updateInterval* 1000);
	this.rightPage= new BezierPatch(config.nx,
					config.nz);
	this.leftPage= new BezierPatch(config.nx,
				       config.nz);

	
	int r, c;
	for(r= 0; r < 4; r++){
	    for(c= 0; c < 4; c++){
		this.rightFinishCtrl_points[r][c]= new Point3d(rightFinishCtrl_pts[r][c]);
		this.rightStartCtrl_points[r][c]= new Point3d(rightStartCtrl_pts[r][c]);
		this.rightIntermediateCtrl_points[r][c]= new Point3d();
		
		this.leftStartCtrl_points[r][c]= new Point3d(leftStartCtrl_pts[r][c]);
		this.leftFinishCtrl_points[r][c]= new Point3d(leftFinishCtrl_pts[r][c]);
		this.leftIntermediateCtrl_points[r][c]= new Point3d();
		
	    }
	}
	
	right_page_num=  bc.sp[0].getPage_num();
	left_page_num=  bc.sp[1].getPage_num();

	//left page image
	if(left_page_num == -1 && bc.coverState == OPEN){
	    if(config.parameters.getfileType() == DJVU){
		leftTextureName= bc.fixedTexName[INSIDE_BOOK_FRONT_COVER];
	    }else{
		leftTextureName= NONE;
	    }
	}
	else if( left_page_num > config.firstPage && bc.coverState == OPEN){
	    leftTextureName= bc.sp[1].getTextureName();
	    leftMaterialName= bc.sp[1].getPageMaterial();//changed
	}

	else if(left_page_num == config.lastPage && bc.coverState == CLOSE){
	    if(config.parameters.getfileType() == DJVU){
		leftTextureName= bc.fixedTexName[BOOK_BACK_COVER];
	    }else{
		leftTextureName= NONE;
	    }
	    
	}
	else{
	    leftTextureName=NONE;

	}

	//right page image
	if(right_page_num > config.lastPage && bc.coverState == OPEN){
	    if(config.parameters.getfileType() == DJVU){
		rightTextureName= bc.fixedTexName[INSIDE_BOOK_BACK_COVER];
	    }else{
		rightTextureName= NONE;
	    }
	}
	else if( right_page_num <= config.lastPage && bc.coverState == OPEN){
	    rightTextureName= bc.sp[0].getTextureName();
	    rightMaterialName= bc.sp[0].getPageMaterial();//changed
	}
	
	else if(right_page_num == config.firstPage && bc.coverState == CLOSE){
	    rightTextureName= bc.fixedTexName[BOOK_FRONT_COVER];
	    
	}
	else{
	    rightTextureName= NONE;
	}
	
		

	if(direction == 1){
	    interpolate(1);
	    Scene.display_mode= Three_D_MODE;
	}else{
	    interpolate(0);
	}
	
	running= true;
    }

    
    public void run()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("Animated.java: run()");
	}
	
	double alpha= 0;
	int step= 1; 
	
	while( step <= totalsteps && !stop){
	    try {
		sleep((int)(msecsUPdateInterval));
	    }catch (InterruptedException e) { }

	    //interpolate and work out intermediate positions of the four corners
	    alpha= (double)step/(double)totalsteps;

	    if(direction == 1){ // changed
		alpha= 1.0 - alpha;
	    }

	    interpolate(alpha);
	    
	    step++;
	}

	running= false;
	


    }

    protected void interpolate(double alpha)
    {
	
	//System.out.println("Animated.java: interpolate() alpha= " + alpha);
	    
	
	int r, c;
	for(r= 0; r < 4; r++){
	    for(c= 0; c < 4; c++){
		
		if(right_page_num > config.lastPage && bc.coverState == CLOSE){
		    
		}else{
		    this.rightIntermediateCtrl_points[r][c].interpolate(rightStartCtrl_points[r][c], 
									rightFinishCtrl_points[r][c],
									alpha);
		}

		if(left_page_num == -1 && bc.coverState == CLOSE){
		    
		}else if(left_page_num == config.lastPage && bc.coverState == CLOSE &&
			 config.parameters.getfileType() == PDF){
		    
		}else{
		    this.leftIntermediateCtrl_points[r][c].interpolate(leftStartCtrl_points[r][c], 
								       leftFinishCtrl_points[r][c],
								       alpha);
		}
	    }
	}

	//reconstruct Bezierpatch
	rightPage.setControlPoints(rightIntermediateCtrl_points);
	rightPage.genPatch();
	leftPage.setControlPoints(leftIntermediateCtrl_points);
	leftPage.genPatch();
    }
    

    public void setStop()
    {
	stop= true;
	
	
	
    }
    
    public boolean isRunning()
    {
	return running;
    }

    /*debug*/
    public void drawEndPages(GLCanvas canvas)
    {
	GLFunc gl = canvas.gl;
	
	float [] blue= {1.f, 1.f, 1.f, 1.f};
	gl.glPushAttrib(GL_ALL_ATTRIB_BITS);
	gl.glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, blue);
	gl.glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, blue);
	gl.glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 1);
	
	gl.glBegin (GL_POINTS);
	
	int r, c;
	/*for(r=0; r<4; r++){
	    for(c=0; c<4; c++){
		gl.glVertex3dv(leftStartCtrl_points[r][c].get3dv());
	    }
	    }*/
	
	
	for(r=0; r<4; r++){
	    for(c=0; c<4; c++){
		gl.glVertex3dv(leftFinishCtrl_points[r][c].get3dv());
	    }
	}

	
	gl.glEnd();
	gl.glPopAttrib();
    }

    public void drawIntermediatePage(GLCanvas canvas,  byte command)
    {
	if(right_page_num > config.lastPage && bc.coverState == CLOSE){
		    
	}else{
	    rightPage.evaluate(canvas, command,  rightTextureName, rightMaterialName ); //changed
	}

	
	if(left_page_num == -1 && bc.coverState == CLOSE){
	
	}else if(left_page_num == config.lastPage && bc.coverState == CLOSE && 
		 config.parameters.getfileType() == PDF){
	    
	}
	else{
	    leftPage.evaluate(canvas, command,  leftTextureName, leftMaterialName); //change
	}
	
    }
    
    
    
}
