package TTPproject;

import java.awt.event.*;

import gl4java.awt.GLCanvas;
import gl4java.GLContext;
import gl4java.GLFunc;
import gl4java.GLUFunc;
import gl4java.GLEnum;


import java.awt.*;
import java.util.Date;

/* Animation
 * getting book to display itself on the given canvas at
 * a specified frame rate (if possible).
 * 
 */


public class Animation extends Thread 
    implements GLEnum, CONSTANTS{

    private ActionListener callback;
    private Scene scene;
    private Configuration config= null;
    private BookConstructor bc;
    private double timeStep; 
    private double angleStep;
    private double t;
    private long msecsDisplayInterval;
    private long turningInterval;//rapid flipping
    private long pausingInterval;//rapid flipping
    private CursorManager cursorManager;

    /*gl4java*/
    private GLContext glj;
    private GLFunc gl;
    private GLUFunc glu;


    //private List anchorList;
    //private int curIndex;
    
    private boolean killed;
    private boolean threadSuspended;
    private boolean stop;
    private boolean swap;
    
    private int delta_x= 0;
    private double numSecsToRun= 3;
    
    
    public Animation(Scene scene, BookConstructor bc, CursorManager cursorManager, 
		     Configuration config,
		     double timeStep, 
		     double displayInterval, 
		     double angleStep,
		     ActionListener callee) {
	this.callback= callee;
	this.scene= scene;
	this.glj= scene.getGLContext();
	this.gl= scene.getGL();
	this.glu= scene.getGLU();

	this.bc= bc;
	this.cursorManager= cursorManager;
	this.config= config;
	this.timeStep= timeStep;
	this.msecsDisplayInterval = (long) (displayInterval * 1000);
	this.turningInterval= (long) (config.getTurningInterval() * 1000);
	this.pausingInterval= (long) (config.getPausingInterval() * 1000);
	this.angleStep= angleStep;
	this.bc= bc;
	
	
	this.killed= false;
	this.threadSuspended= false;
	this.stop= false;
	this.swap= false;
    }

    public void setNumSecsToRun(double numSecsToRun)
    {
	this.numSecsToRun= numSecsToRun;
    }

    /*
      At the moment, which objects (e.g. Either  Front/Back cover or Pages) are selected depends 
      on the 'stage' (interval)
      
      For example, stage 0 the only moving object is the front cover
    */
    public void run()   {
	
	killed= false;
  	//scene.repaint();
	
	//Date d = new Date();
        //double nextSimTimeToDisplay = displayInterval;
        //long nextRealTimeToDisplay = d.getTime() + msecsDisplayInterval;
	//setPriority(Thread.NORM_PRIORITY - 2);  // Lower our priority so GUI stays active
	if(Ttp.commentslevel >= 2){
	    System.out.println("Animation: Start running");
	}
	
	//System.out.println("Animation: current cover angle " + bc.getFrontCoverTurningAngle());
	//int selected_name= bc.getSelectedName();

	if( scene.selected_name == config.FRONT_COVER){
	    if(Ttp.commentslevel >= 3){
		System.out.println("Animation: front cover selected");
	    }
	    /*
	      front cover
	    */
	    if(bc.getCoverState() == CLOSE ){
		if(Ttp.commentslevel >= 3){
		    System.out.println("Animation: cover is closed");
		}
		if(bc.isFrontCoverFacingUP()){
		    while( !bc.isCloseFrontCover_Rest()){
			//System.out.println("angleStep: " + angleStep);
			try {
			    sleep((int)(msecsDisplayInterval));
			    //suspend 
			    synchronized(this) {
				while (threadSuspended){
				    if(Ttp.commentslevel >= 3){
					System.out.println("Animation.java: run() getSuspended" );
				    }
				    wait();
				}
			    }
			}catch (InterruptedException e) { }
			
			bc.rotateCloseFrontCover(angleStep);
		    }
		    
		    if(bc.getFrontCoverTurningAngle() < config.HALF_WAY){
			bc.deleteCurTexName();
			bc.reset_StaticPages();
			bc.pageReconstruction2();
			callback.actionPerformed(new ActionEvent(this, 0, "CLOSE_COVER_UNSELECTED_CLOSE"));
		    }
		    else{
			bc.setCoverState(OPEN);
			//bc.pageReconstruction();
			bc.pageReconstruction2();
			callback.actionPerformed(new ActionEvent(this, 0, "CLOSE_COVER_UNSELECTED_OPEN"));
		    }
		
		    
		}else{
		    //front cover facing down
		    
		    //restore
		    while( !bc.isFrontCover_TurnOver()){
			if(Ttp.commentslevel >= 3){
			    System.out.println("angleStep: " + angleStep);
			}
			try {
			    sleep((int)(msecsDisplayInterval));
			    //suspend 
			    synchronized(this) {
				while (threadSuspended){
				    if(Ttp.commentslevel >= 3){
					System.out.println("Animation.java: run() getSuspended" );
				    }
				    wait();
				}
			    }
			}catch (InterruptedException e) { }
			
			bc.flipFrontCover(angleStep);
		    }	

		    if(bc.getFrontCoverTurningAngle() > config.HALF_WAY){
			
		    }else{
			bc.reset_StaticPages();
			bc.pageReconstruction2();
		    }
		    

		}
		
		
	    }else{//is open
		if(Ttp.commentslevel >= 3){
		    System.out.println("Animation: cover is opened");
		}
		
		//rotate clockwisely until closed
		while(! bc.isOpenFrontCover_Rest()){
		    try {
			sleep((int)(msecsDisplayInterval));
			//suspend 
			synchronized(this) {
			    while (threadSuspended){
				if(Ttp.commentslevel >= 3){
				    System.out.println("Animation.java: run() getSuspended" );
				}
				wait();
			    }
			}
		    }catch (InterruptedException e) { }
		    
		    bc.rotateOpenFrontCover(angleStep);
		}
		
		if(bc.getFrontCoverTurningAngle() < config.HALF_WAY){
		    bc.setCoverState(CLOSE);  // !!!!!!!!
		    bc.reset_StaticPages();
		    callback.actionPerformed(new ActionEvent(this, 0, "OPEN_COVER_UNSELECTED_CLOSE"));
		    bc.pageReconstruction2();
		    bc.deconstructBundlePages();//disconstruct pagebundles
		}else{
		    callback.actionPerformed(new ActionEvent(this, 0, "OPEN_COVER_UNSELECTED_OPEN"));
		}
		
	    }
	   
	    scene.selection_mode= UNSELECTED;
	    scene.turning_mode= NOT_ANIMATING;// reset the restriction
	    scene.setCursor(cursorManager.getCursor(IDLE));
	}
	
	else if(scene.selected_name == config.BACK_COVER){
	    if(Ttp.commentslevel >= 3){
		System.out.println("Animation: back cover selected");
	    }
	    if(bc.getCoverState() == CLOSE){
		if(Ttp.commentslevel >= 3){
		    System.out.println("Animation: cover is closed");
		}

		if(bc.isFrontCoverFacingDOWN()){//backcover facing up
		    if(Ttp.commentslevel >= 3){
			System.out.println("\tBack cover facing UP");
		    }
		    if(Ttp.commentslevel >= 3){
			System.out.println("isBackCoverCompletely close again");
		    }
		    while( !bc.isCloseBackCover_Rest() ){
			try {
			    sleep((int)(msecsDisplayInterval));
			    //suspend 
			    synchronized(this) {
				while (threadSuspended){
				    if(Ttp.commentslevel >= 3){
					System.out.println("Animation.java: run() getSuspended" );
				    }
				    wait();
				}
			    }
			}catch (InterruptedException e) { }
			
			bc.rotateCloseBackCover(angleStep);
		    }
			
		    if(bc.getBackCoverTurningAngle() > config.HALF_WAY ){
			bc.deleteCurTexName();
			bc.reset_StaticPages();
			bc.pageReconstruction2();
			callback.actionPerformed(new ActionEvent(this, 0, "CLOSE_COVER_UNSELECTED_CLOSE"));
		    }else{
			bc.setCoverState(OPEN);
			//bc.pageReconstruction();
			bc.pageReconstruction2();
			callback.actionPerformed(new ActionEvent(this, 0, "CLOSE_COVER_UNSELECTED_OPEN"));
		    }
		    
		}else{
		    //backcover facing down
		    if(Ttp.commentslevel >= 3){
			System.out.println("\tBack cover facing DOWN");
		    }
		    
		    while( !bc.isBackCover_TurnOver() ){
			try {
			    sleep((int)(msecsDisplayInterval));
			    //suspend 
			    synchronized(this) {
				while (threadSuspended){
				    if(Ttp.commentslevel >= 3){
					System.out.println("Animation.java: run() getSuspended" );
				    }
				    wait();
				}
			    }
			}catch (InterruptedException e) { }
			
			bc.flipBackCover(angleStep);
		    }
		    if(bc.getBackCoverTurningAngle() < config.HALF_WAY ){

		    }else{
			bc.reset_StaticPages();
			bc.pageReconstruction2();
		    }
		}
	    }else{
		if(Ttp.commentslevel >= 3){
		    System.out.println("Animation: cover is opened");
		}
		while(! bc.isOpenBackCover_Rest()){
		    try {
			sleep((int)(msecsDisplayInterval));
			//suspend 
			synchronized(this) {
			    while (threadSuspended){
				if(Ttp.commentslevel >= 3){
				    System.out.println("Animation.java: run() getSuspended" );
				}
				wait();
			    }
			}
		    }catch (InterruptedException e) { }
		    
		    bc.rotateOpenBackCover(angleStep);
		}
		
		if(bc.getBackCoverTurningAngle() > config.HALF_WAY ){
		    bc.setCoverState(CLOSE);
		    bc.reset_StaticPages();
		    callback.actionPerformed(new ActionEvent(this, 0, "OPEN_COVER_UNSELECTED_CLOSE"));
		    bc.pageReconstruction2();
		    bc.deconstructBundlePages();//disconstruct pagebundles
		}else{
		    callback.actionPerformed(new ActionEvent(this, 0, "OPEN_COVER_UNSELECTED_OPEN"));
		}
	    }
	    
	    scene.selection_mode= UNSELECTED;
	    scene.turning_mode= NOT_ANIMATING;// reset the restriction
	    scene.setCursor(cursorManager.getCursor(IDLE));

	}else if(scene.selected_name == config.RIGHT_EDGE || 
		 scene.selected_name == config.LEFT_EDGE ){
	    
	    if(bc.getCoverState() == CLOSE){
		if(Ttp.commentslevel >= 3){
		    System.out.println("\tCover closed");
		}
		
		if(scene.selected_name == config.RIGHT_EDGE){
		    if(Ttp.commentslevel >= 3){
			System.out.println("\tRIGHT_EDGE");
		    }
		    if(bc.getFrontCoverTurningAngle() < config.HALF_WAY ){
			// closed
			while( !bc.isFrontCoverCompletelyCloseAgain() && !killed ){
			    if(Ttp.commentslevel >= 3){
				System.out.println("angleStep: " + angleStep);
			    }
			    bc.rotateSpine_RightEdgeSelected(angleStep);
			    
			    try { sleep(msecsDisplayInterval ); }       // sleep msecsDisplayInterval
			    catch (InterruptedException e) {};
			    //scene.repaint(); 
			}	
			
			if(!killed){
			    //bc.pageReconstruction();
			    //bc.reset_theTwoStaticPages();
			    //bc.deleteCurTexName();
			    bc.reset_StaticPages();
			    bc.pageReconstruction2();
			    callback.actionPerformed(new ActionEvent(this, 0, "CLOSE_EDGE_UNSELECTED_CLOSE"));
			}
		    }
		    else{
			//open
			while( !bc.isFrontCoverCompletelyOpen() && !killed ){
			    if(Ttp.commentslevel >= 3){
				System.out.println("angleStep: " + -angleStep);
			    }
			    bc.rotateSpine_RightEdgeSelected(-angleStep);
			    
			    try { sleep(msecsDisplayInterval ); }       // sleep msecsDisplayInterval
			    catch (InterruptedException e) {};
			    //scene.repaint();
			}
		    
			if(!killed){
			    bc.setCoverState(OPEN);
			    bc.updatePapers();
			    bc.pageReconstruction2();
			    callback.actionPerformed(new ActionEvent(this, 0, "CLOSE_EDGE_UNSELECTED_OPEN"));
			    //bc.startinitializingPageContent();
			    //bc.bufferingOption= WINDOWING;
			}
		    }
		}else{
		    if(Ttp.commentslevel >= 3){
			System.out.println("\tLEFT_EDGE");
		    }
		    if(bc.getBackCoverTurningAngle() > config.HALF_WAY ){
			if(Ttp.commentslevel >= 3){
			    System.out.println("isBackCoverCompletely close again");
			}
			while( !bc.isBackCoverCompletelyCloseAgain() && !killed ){
			    bc.rotateSpine_LeftEdgeSelected(-angleStep); 
			    
			    try { Thread.sleep(msecsDisplayInterval ); }       // sleep msecsDisplayInterval
			    catch (InterruptedException e) {};
			    //scene.repaint(); 
			}

			if(!killed){
			    //bc.pageReconstruction();
			    //bc.reset_theTwoStaticPages();
			    //bc.deleteCurTexName();
			    bc.reset_StaticPages();
			    bc.pageReconstruction2();
			    callback.actionPerformed(new ActionEvent(this, 0, "CLOSE_EDGE_UNSELECTED_CLOSE"));
			}
			
		    }else{
			if(Ttp.commentslevel >= 3){
			    System.out.println("isBackCoverCompletely open ");
			}
			while( !bc.isBackCoverCompletelyOpen() && !killed ){
			    bc.rotateSpine_LeftEdgeSelected(angleStep); 

			    try { Thread.sleep(msecsDisplayInterval ); }       // sleep msecsDisplayInterval
			    catch (InterruptedException e) {};
			    //scene.repaint(); 
			}
			
			if(!killed){
			    bc.setCoverState(OPEN);
			    //bc.deleteCurTexName();
			    //bc.pageReconstruction();
			    bc.updatePapers();
			    bc.pageReconstruction2();
			    callback.actionPerformed(new ActionEvent(this, 0, "CLOSE_EDGE_UNSELECTED_OPEN"));
			}
		    
		    }
		}
	    }else{//coverState == OPEN
		if(bc.getSelectedSide() == RIGHT){
		    System.out.println("\tRight hand page has been selected");
		    if(bc.getPageTurningAngle() < config.HALF_WAY ){
			while( !bc.isRightPage_rest_on_Right()  && !killed){
			    if(Ttp.commentslevel >= 3){
				System.out.println("angleStep: " + angleStep);
			    }
			    bc.rotateSpine(angleStep);
			
			    try { Thread.sleep(msecsDisplayInterval ); }       // sleep msecsDisplayInterval
			    catch (InterruptedException e) {};
			    //scene.repaint();  
			}
			if(!killed){
			    bc.step_back_step_SelectedOneStaticPage();//bc.jump_back_step_SelectedOneStaticPage();
			    bc.pageReconstruction2();//the new model
			    callback.actionPerformed(new ActionEvent(this, 0, "OPEN_EDGE_UNSELECTED_RESTORE"));
			}
			
		    }else{
			while( !bc.isRightPage_rest_on_Left() && !killed ){
			    if(Ttp.commentslevel >= 3){
				System.out.println("angleStep: " + -angleStep);
			    }
			    bc.rotateSpine(-angleStep);
			    
			    try { Thread.sleep(msecsDisplayInterval ); }       // sleep msecsDisplayInterval
			    catch (InterruptedException e) {};
			    //scene.repaint();  
			}
			
			if(!killed){
			    bc.jump_theOtherStaticPage();//bc.jump_theOtherStaticPage()
			    callback.actionPerformed(new ActionEvent(this, 0, "OPEN_EDGE_UNSELECTED_UPDATE"));
			    bc.updatePapers();
			    bc.update_StaticPages();
			    bc.pageReconstruction2();
			    
			    
			}
		    }
		}else{
		    if(Ttp.commentslevel >= 3){
			System.out.println("\tLeft hand page has been selected");
		    }
		    
		    if(bc.getPageTurningAngle() < -config.HALF_WAY ){
			while( !bc.isLeftPage_rest_on_Right() && !killed ){
			    if(Ttp.commentslevel >= 3){
				System.out.println("angleStep: " + angleStep);
			    }
			    bc.rotateSpine(angleStep);
			    
			    try { Thread.sleep(msecsDisplayInterval ); }       // sleep msecsDisplayInterval
			    catch (InterruptedException e) {};
			    //scene.repaint(); 
			}
			
			if(!killed){
			    bc.jump_theOtherStaticPage();
			    callback.actionPerformed(new ActionEvent(this, 0, "OPEN_EDGE_UNSELECTED_UPDATE"));
			    bc.updatePapers();
			    bc.update_StaticPages();
			    bc.pageReconstruction2();
			    
			}
		    }else{
			
			while( !bc.isLeftPage_rest_on_Left() && !killed ){
			    if(Ttp.commentslevel >= 3){
				System.out.println("angleStep: " + angleStep);
			    }
			    bc.rotateSpine(-angleStep);
			    
			    try { Thread.sleep(msecsDisplayInterval ); }       // sleep msecsDisplayInterval
			    catch (InterruptedException e) {};
			    //scene.repaint(); 
			}

			if(!killed ){
			    bc.step_back_step_SelectedOneStaticPage();
			    bc.pageReconstruction2();//the new model
			    callback.actionPerformed(new ActionEvent(this, 0, "OPEN_EDGE_UNSELECTED_RESTORE"));
			    
			}
		    }
		}
	    }
	    
	    if(!killed){// killed should be interpreted as suspended
		if(Ttp.commentslevel >= 3){
		    System.out.println("\t Animation.java: run() selection_mode == UNSELECTED");
		}
		//if texture still rendering
		scene.selection_mode= UNSELECTED;
		scene.turning_mode= NOT_ANIMATING;// reset the restriction
		scene.setCursor(cursorManager.getCursor(IDLE));
		
	    }
	    else if(killed)
		scene.turning_mode= NOT_ANIMATING;// release the turning restriction
	
	}	
	else if(scene.selected_name == config.FORWARD_BUTTON || 
		scene.selected_name == config.BACKWARD_BUTTON 
		){
	    //it is called only when right animation buttons has been selected
	    //be clear Animation is responsible for turning the pages
	    //but structurally turning the first chuck and (turning the last chunk)
	    //is different from turning the chunks inbetween because the way the
	    //book is constructed differently (call different functions)
	    
	    //other condition like is front cover facing up or down
	    java.util.List anchorList;
	    int startIndex, pageDifference;
	    int curPage;
	    

	    for(;!stop;){
		//if forward {
		if(scene.selected_name == config.FORWARD_BUTTON){
		    anchorList= scene.getRightAnchorList();
		    startIndex= scene.getCurRightIndex();
		    
		    if(startIndex == NONE){
			stop= true;
			break;
		    }
		    
		    int nextPage; // anchored page
		    curPage= ((Integer)anchorList.get(startIndex)).intValue();
		    
		    if(curPage != bc.sp[0].getPage_num()){ // does not coincide
			curPage= bc.sp[0].getPage_num();
		    }else{
			//does concide
			if(curPage == 0){// the first page
			    
			}else{
			    startIndex += 1;
			}
		    }

		    if(curPage >= config.lastPage){ 
			stop= true;
		    }
		    
		    if(startIndex == (anchorList.size())){
			stop= true;
			break;
		    }
		    
		    
		    for(int index= startIndex; 
			index < anchorList.size() && !stop && !swap; 
			index++){
			if(Ttp.commentslevel >= 3){
			    System.out.println("Animation.java:  run(): forwarding ");
			}
			
			
			
			nextPage= ((Integer)anchorList.get(index)).intValue();// read the next page
			if(Ttp.commentslevel >= 3){
			    System.out.println("Animation.java:  run(): right page= " + bc.sp[0].getPage_num());
			    System.out.println("Animation.java:  run(): left page= " + bc.sp[1].getPage_num());
			    System.out.println("Animation.java: run() curPage= " + curPage);
			    System.out.println("Animation.java: run() nextPage= " + nextPage);
			}
			
			scene.setCurRightIndex(index);// update Cur right index
			scene.setCurLeftIndex(index);// update Cur left index
			pageDifference= nextPage - curPage;// positive value
			
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("Animation.java: initializeStartingAnchorPosition() pageDifference= " + 
					       pageDifference);
			}
			
			if(bc.getCoverState() == CLOSE){
			    //book is closed when startPage is the first page and the last page
			    
			    if(bc.isFrontCoverFacingUP()){
				//we should open to the first anchor page
				if(pageDifference == 0){
				    //turning the front cover
				    bc.setSelectedFrontCover();
				    bc.jump_SelectedOneStaticPage();
				    bc.jump_theOtherStaticPage();

				    //double numSecsToRun= 3;
				    double secTurningInterval= config.getTurningInterval(); 
				    double secs= 0;
				    double t;

				    
				    while(secs < numSecsToRun ){
					
					secs += secTurningInterval;
					t= secs/numSecsToRun;
					bc.stepFrontCoverState(t);
					
					try { sleep(turningInterval ); }       // in  msecs
					catch (InterruptedException e) {};
				    }
				    				    
				    bc.setCoverState(OPEN);
				    bc.pageReconstruction2();
				    
				}else{
				    //turning initialization
				    bc.setSelectedRightEdge(pageDifference);
				    bc.jump_SelectedOneStaticPage();
				    bc.jump_theOtherStaticPage();
				    
				    //double numSecsToRun= 3;
				    double secTurningInterval= config.getTurningInterval(); 
				    double secs= 0;
				    double t;

				    while(secs < numSecsToRun ){
					secs += secTurningInterval;
					t= secs/numSecsToRun;
					bc.stepSpine_RightEdgeState(t);
					
					try { sleep(turningInterval ); }       // in  msecs
					catch (InterruptedException e) {};
				    }
				    
				    
				    //perform turning
				    /*while( !bc.isFrontCoverCompletelyOpen()){
					//System.out.println("angleStep: " + -angleStep);
					bc.rotateSpine_RightEdgeSelected(-config.flippingAngleStep);
					
					try { sleep(turningInterval); }       // sleep msecsDisplayInterval
					catch (InterruptedException e) {};
					//scene.repaint();
				    }*/
				    
				    //turning termination
				    bc.setCoverState(OPEN);
				    bc.updatePapers();
				    bc.pageReconstruction2();
				}
			    }
			}
			
			else if(bc.getCoverState() == OPEN){

			    if(pageDifference == 0){//from the first page 
				
			    }else{
				if(pageDifference == 2 ){
				    Scene.turnNumofPages= 2;
				    if(Ttp.commentslevel >= 3){
					System.out.println("Animation.java: 2 pages");
				    }
				    bc.setSelectedPage3(scene.selected_name);
				    
				    config.setRegion(1);
				    bc.setDynamicPage();/*get ready for interpolation, set initial page rotation angle
							  stop rotation angle*/
				    
				    //scene.selected_name= bc.getSelectedpage();//****
				    Scene.readytodisplayMovingPage= true;
				    bc.jump_SelectedOneStaticPage();// update static page's page number and texturename
				    
				    
				    //double numSecsToRun= 3;
				    double secTurningInterval= config.getTurningInterval(); 
				    double secs= 0;
				    double t;
				    
				    while(secs < numSecsToRun ){
					secs += secTurningInterval;
					t= secs/numSecsToRun;
					bc.stepPage_State(t);
					
					try { sleep(turningInterval ); }       // in  msecs
					catch (InterruptedException e) {};
				    }
					
				    
				    bc.jump_theOtherStaticPage();
				    bc.updatePapers();
				    bc.update_StaticPages();
					bc.pageReconstruction2();
					
					//reset
					Scene.readytodisplayMovingPage= false;
					Scene.turnNumofPages= 0;
					
					//reset selected_name***
					if(swap){
					    if(delta_x < 0){ // Negative
					    scene.selected_name =config.FORWARD_BUTTON;//forwarding
					    }else{           //Positive
						scene.selected_name =config.BACKWARD_BUTTON;
					    }
					}else{
					    scene.selected_name= config.FORWARD_BUTTON;
					}
					

				    }else{//
					//turning initialization
					Scene.turnNumofPages= pageDifference;
					Scene.readytodisplayMovingPage= true;
					bc.setSelectedEdge(config.RIGHT_EDGE, pageDifference);
					bc.setBundlePages();
					bc.jump_SelectedOneStaticPage();
					
					
					//double numSecsToRun= 3;
					double secTurningInterval= config.getTurningInterval(); 
					double secs= 0;
					double t;
					
					while(secs < numSecsToRun ){
					    secs += secTurningInterval;
					    t= secs/numSecsToRun;
					    bc.stepSpine_State(t);
					    
					    try { sleep(turningInterval ); }       // in  msecs
					    catch (InterruptedException e) {};
					}
				    
					
					//perform turning
					/*while( !bc.isRightPage_rest_on_Left()){
					//System.out.println("angleStep: " + -angleStep);
					bc.rotateSpine(-config.flippingAngleStep);
					
					try { sleep(turningInterval ); }
					catch (InterruptedException e) {};
					//scene.repaint();  
					}*/
					
					//turning termination
					bc.jump_theOtherStaticPage();
					//bc.step_theOtherStaticPage2();//bc.jump_theOtherStaticPage()
					bc.updatePapers();
					bc.update_StaticPages();
					bc.pageReconstruction2();
					bc.deconstructBundlePages();//disconstruct pagebundles
					
					//reset
					Scene.readytodisplayMovingPage= false;
					Scene.turnNumofPages= 0;
					
					//reset selected_name
					if(swap){
					    if(delta_x < 0){ // Negative
						scene.selected_name =config.FORWARD_BUTTON;//forwarding
					    }else{           //Positive
						scene.selected_name =config.BACKWARD_BUTTON;
					    }
					}
				    }
			    }
			    
			}
			curPage= nextPage;
			//pausing for a few second 
			try {
			    sleep(pausingInterval);
			}catch (InterruptedException e) { }
		    }
		}
		    
		
		//else backwared
		else if(scene.selected_name == config.BACKWARD_BUTTON){
		    anchorList= scene.getLeftAnchorList();
		    startIndex= scene.getCurLeftIndex();
		    
		    if(startIndex == NONE){
			stop= true;
			break;
		    }

		    int previousPage; 
		    curPage= ((Integer)anchorList.get(startIndex)).intValue();
		    
		    if(curPage != bc.sp[1].getPage_num()){ // does not coincide
			curPage= bc.sp[1].getPage_num();
		    }else{
			//does coninde
			if(curPage == config.lastPage){// the last page
			    
			}else{
			    startIndex -= 1;
			}
		    }
		    
		    if(curPage <= config.firstPage){ 
			stop= true;
		    }

		    if(startIndex < 0){
			stop= true;
			break;
		    }
		    
		    for(int index= startIndex;  
			index >= 0 && !stop && !swap; 
			index--){
			if(Ttp.commentslevel >= 3){
			    System.out.println("Animation.java:  run(): Backwarding ");
			}
			previousPage= ((Integer)anchorList.get(index)).intValue();

			if(Ttp.commentslevel >= 3){
			    System.out.println("Animation.java: initializeStartingAnchorPosition() curPage= " + 
					       curPage);
			    System.out.println("Animation.java: initializeStartingAnchorPosition() previousPage= " + 
					       previousPage);
			}

			scene.setCurLeftIndex(index);// update Cur left index
			scene.setCurRightIndex(index);// update Cur right index
			pageDifference= curPage - previousPage;// positive value
			System.out.println("Animation.java: pagedifference=" + pageDifference );
			
			
			if(bc.getCoverState() == CLOSE){
			    if(bc.isFrontCoverFacingDOWN()){
				if(pageDifference == 0){
				    //turning the back cover
				    //turning initialization
				    bc.setSelectedBackCover();
				    bc.jump_SelectedOneStaticPage();
				    bc.jump_theOtherStaticPage();
				    
				    //double numSecsToRun= 3;
				    double secTurningInterval= config.getTurningInterval(); 
				    double secs= 0;
				    double t;

				    while(secs < numSecsToRun ){
					secs += secTurningInterval;
					t= secs/numSecsToRun;
					bc.stepBackCoverState(t);
					
					try { sleep(turningInterval ); }       // in  msecs
					catch (InterruptedException e) {};
				    }
			    
				    //perform turning
				    /*while( !bc.isBackCoverCompletelyOpen()){
					bc.rotateBackCover(config.flippingAngleStep);
				
					try { sleep(turningInterval ); }
					catch (InterruptedException e) {};
					}*/
				    
				    //turning termination
				    bc.setCoverState(OPEN);
				    bc.pageReconstruction2();
				    
				}else{
				    //jump into any points of the book
				    //may be this will never perform
				    //turning initialization
				    bc.setSelectedLeftEdge(pageDifference);
				    bc.jump_SelectedOneStaticPage();
				    bc.jump_theOtherStaticPage();

				    //double numSecsToRun= 3;
				    double secTurningInterval= config.getTurningInterval(); 
				    double secs= 0;
				    double t;

				    while(secs < numSecsToRun ){
					secs += secTurningInterval;
					t= secs/numSecsToRun;
					bc.stepSpine_LeftEdgeState(t);

					
					try { sleep(turningInterval ); }       // in  msecs
					catch (InterruptedException e) {};
				    }
			    
			    
				    //perform turning
				    /*while( !bc.isBackCoverCompletelyOpen()){
					bc.rotateSpine_LeftEdgeSelected(config.flippingAngleStep); 
					
					try { sleep(turningInterval ); }       // sleep msecsDisplayInterval
					catch (InterruptedException e) {};
					}*/
				    
				    //turning termination
				    bc.setCoverState(OPEN);
				    bc.updatePapers();
				    bc.pageReconstruction2();
				}
				
			    }
			}
			
			else if(bc.getCoverState() == OPEN){
			    if(pageDifference == 0){//from the last page 
				
			    }else{
				if(pageDifference == 2 ){
				    Scene.turnNumofPages= 2;
				    if(Ttp.commentslevel >= 3){
					System.out.println("Animation.java: 2 pages");
				    }
				    
				    bc.setSelectedPage3(scene.selected_name);
				    
				    config.setRegion(1);
				    bc.setDynamicPage();/*get ready for interpolation, set initial page rotation angle
							  stop rotation angle*/
				    //scene.selected_name= bc.getSelectedpage();*****
				    Scene.readytodisplayMovingPage= true;
				    bc.jump_SelectedOneStaticPage();// update static page's page number and texturename

				    //double numSecsToRun= 3;
				    double secTurningInterval= config.getTurningInterval(); 
				    double secs= 0;
				    double t;

				    while(secs < numSecsToRun ){
					secs += secTurningInterval;
					t= secs/numSecsToRun;
					bc.stepPage_State(t);

					
					try { sleep(turningInterval ); }       // in  msecs
					catch (InterruptedException e) {};
				    }
				    
				    bc.jump_theOtherStaticPage();
				    bc.updatePapers();
				    bc.update_StaticPages();
				    bc.pageReconstruction2();

				    //reset
				    Scene.readytodisplayMovingPage= false;
				    Scene.turnNumofPages= 0;

				    //reset selected_name
				    if(swap){
					if(delta_x < 0){ // Negative
					    scene.selected_name =config.FORWARD_BUTTON;//forwarding
					}else{           //Positive
					    scene.selected_name =config.BACKWARD_BUTTON;
					}
				    }else{
					scene.selected_name= config.BACKWARD_BUTTON;
				    }
				}else{
				    //turning initialization
				    
				    Scene.turnNumofPages= pageDifference;
				    Scene.readytodisplayMovingPage= true;
				    bc.setSelectedEdge(config.LEFT_EDGE, pageDifference);
				    bc.setBundlePages();
				    bc.jump_SelectedOneStaticPage();
				     
				    //double numSecsToRun= 3;
				    double secTurningInterval= config.getTurningInterval(); 
				    double secs= 0;
				    double t;

				    while(secs < numSecsToRun ){
					secs += secTurningInterval;
					t= secs/numSecsToRun;
					bc.stepSpine_State(t);

					
					try { sleep(turningInterval ); }       // in  msecs
					catch (InterruptedException e) {};
				    }
				
				    //perform turning
				    /*while( !bc.isLeftPage_rest_on_Right()){
					bc.rotateSpine(config.flippingAngleStep);
					
					try { sleep(turningInterval); }       // sleep msecsDisplayInterval
					catch (InterruptedException e) {};
					}*/
				    
				    //turning termination
				    bc.jump_theOtherStaticPage();
				    bc.updatePapers();
				    bc.update_StaticPages();
				    bc.pageReconstruction2();
				    bc.deconstructBundlePages();

				    //reset
				    Scene.readytodisplayMovingPage= false;
				    Scene.turnNumofPages= 0;

				    //reset selected_name
				    if(swap){
					if(delta_x < 0){ // Negative
					    scene.selected_name =config.FORWARD_BUTTON;//forwarding
					}else{           //Positive
					    scene.selected_name =config.BACKWARD_BUTTON;
					}
				    }
				    
				}
			    }
			}
			
			curPage= previousPage;
			//pausing for a few second 
			try {
			    sleep(pausingInterval);
			    
			}catch (InterruptedException e) { }
		    }
		}
		
		swap= false;
		//if reach the end
		// stop=true;
	    }

	    scene.selection_mode= UNSELECTED;
	    scene.turning_mode= NOT_ANIMATING;// reset the restriction
	    
	    scene.setCursor(cursorManager.getCursor(IDLE));
	    callback.actionPerformed(new ActionEvent(this, 0, "STOP_FLIPPING"));
	}

	else if(scene.selected_name == config.TURN_LEFT || 
		scene.selected_name == config.TURN_RIGHT ){
	    System.out.println("Start animating single page turn");
	    int left_page= 0;
	    int right_page= 0;

	    for(;!stop;){
		left_page= bc.getLeftPageNum();
		right_page= bc.getRightPageNum();

		if(scene.selected_name == config.TURN_RIGHT){
		    if( bc.getCoverState() == CLOSE && left_page == config.lastPage){
			stop=true;
		    }else{
			if(bc.getCoverState() == CLOSE){
			    //turning the front cover
			    bc.setSelectedFrontCover();
			    bc.jump_SelectedOneStaticPage();
			    bc.jump_theOtherStaticPage();
			    
			    //double numSecsToRun= 3;
			    double secTurningInterval= config.getTurningInterval(); 
			    double secs= 0;
			    double t;
			    
			    while(secs < numSecsToRun ){
				secs += secTurningInterval;
				t= secs/numSecsToRun;
				bc.stepFrontCoverState(t);
				
				try { sleep(turningInterval ); }       // in  msecs
				catch (InterruptedException e) {};
			    }
			    
			    bc.setCoverState(OPEN);
			    bc.pageReconstruction2();
			}else{ 
			    //open
			    if( left_page == config.lastPage){
				//I want to close the back cover
				bc.setSelectedBackCover();
				double secTurningInterval= config.getTurningInterval(); 
				double secs= 0;
				double t;
				
				while(secs < numSecsToRun ){
				    secs += secTurningInterval;
				    t= secs/numSecsToRun;
				    bc.stepBackCoverState(t);
				    
				    try { sleep(turningInterval ); }       // in  msecs
				    catch (InterruptedException e) {};
				}
				
				//turning termination
				bc.setCoverState(CLOSE);
				bc.pageReconstruction2();
			    }else{
				//turning the internal page
				bc.setSelectedPage3(scene.selected_name);
				config.setRegion(1); //random ???
				bc.setDynamicPage();/*get ready for interpolation, set initial page rotation angle
						      stop rotation angle*/
				bc.stepPage_State(0);
				Scene.readytodisplayMovingPage= true;
				bc.jump_SelectedOneStaticPage();// update static page's page number and texturename
				
				//scene.selected_name= bc.getSelectedpage();
				double secTurningInterval= config.getTurningInterval(); 
				double secs= 0;
				double t;
				
				while(secs < numSecsToRun ){
				    secs += secTurningInterval;
				    t= secs/numSecsToRun;
				    bc.stepPage_State(t);
				    
				    try { sleep(turningInterval ); }       // in  msecs
				    catch (InterruptedException e) {};
				}
				
				bc.jump_theOtherStaticPage();
				bc.updatePapers();
				bc.update_StaticPages();
				bc.pageReconstruction2();
				
				//reset
				Scene.readytodisplayMovingPage= false;
			    }
			}
		    }
		}else{ //TURN_LEFT
		    if( bc.getCoverState() == CLOSE && right_page == config.firstPage){
			stop=true;
		    }else{
			if(bc.getCoverState() == CLOSE){
			    bc.setSelectedBackCover();
			    
			    double secTurningInterval= config.getTurningInterval(); 
			    double secs= 0;
			    double t;
			    
			    while(secs < numSecsToRun ){
				secs += secTurningInterval;
				t= secs/numSecsToRun;
				bc.stepBackCoverState(t);
				
				try { sleep(turningInterval ); }       // in  msecs
				catch (InterruptedException e) {};
			    }
			    //rotate clockwisely until closed
			    
			    bc.setCoverState(OPEN);
			    bc.pageReconstruction2();
			    
			}else{
			    //open
			    if( right_page == config.firstPage){
				//I want to close the back cover
				bc.setSelectedFrontCover();
				double secTurningInterval= config.getTurningInterval(); 
				double secs= 0;
				double t;
				
				while(secs < numSecsToRun ){
				    secs += secTurningInterval;
				    t= secs/numSecsToRun;
				    bc.stepFrontCoverState(t);
				    
				    try { sleep(turningInterval ); }       // in  msecs
				    catch (InterruptedException e) {};
				}
				
				//turning termination
				bc.setCoverState(CLOSE);
				bc.pageReconstruction2();
			    }else{
				//turning the internal page
				bc.setSelectedPage3(scene.selected_name);
				config.setRegion(1); //random ???
				bc.setDynamicPage();/*get ready for interpolation, set initial page rotation angle
						      stop rotation angle*/
				bc.stepPage_State(0);
				Scene.readytodisplayMovingPage= true;
				bc.jump_SelectedOneStaticPage();// update static page's page number and texturename
				
				//scene.selected_name= bc.getSelectedpage();
				double secTurningInterval= config.getTurningInterval(); 
				double secs= 0;
				double t;
				
				while(secs < numSecsToRun ){
				    secs += secTurningInterval;
				    t= secs/numSecsToRun;
				    bc.stepPage_State(t);
				    
				    try { sleep(turningInterval ); }       // in  msecs
				    catch (InterruptedException e) {};
				}
				
				bc.jump_theOtherStaticPage();
				bc.updatePapers();
				bc.update_StaticPages();
				bc.pageReconstruction2();
				
				//reset
				Scene.readytodisplayMovingPage= false;
			    }
			}

		    }

		}

	    }

		      

	    //if(scene.selected_name == config.TURN_RIGHT){
	    //	System.out.println("!!!Right page");
	    /*
	    if(bc.getCoverState() == CLOSE){
		    if(scene.selected_name == config.TURN_RIGHT){
			//turning the front cover
			bc.setSelectedFrontCover();
			bc.jump_SelectedOneStaticPage();
			bc.jump_theOtherStaticPage();
			
			//double numSecsToRun= 3;
			double secTurningInterval= config.getTurningInterval(); 
			double secs= 0;
			double t;

			while(secs < numSecsToRun ){
			    secs += secTurningInterval;
			    t= secs/numSecsToRun;
			    bc.stepFrontCoverState(t);
			    
			    try { sleep(turningInterval ); }       // in  msecs
			    catch (InterruptedException e) {};
			}
			
			bc.setCoverState(OPEN);
			bc.pageReconstruction2();
		    }else{
			if(bc.isFrontCoverFacingDOWN()){ //ensure
			    //turning the back cover
			    //turning initialization
			    bc.setSelectedBackCover();
			    bc.jump_SelectedOneStaticPage();
			    bc.jump_theOtherStaticPage();
			    
			    //double numSecsToRun= 3;
			    double secTurningInterval= config.getTurningInterval(); 
			    double secs= 0;
			    double t;
			    
			    while(secs < numSecsToRun ){
				secs += secTurningInterval;
				t= secs/numSecsToRun;
				bc.stepBackCoverState(t);
				
				try { sleep(turningInterval ); }       // in  msecs
				catch (InterruptedException e) {};
			    }
			    
			    
			    
			    //turning termination
			    bc.setCoverState(OPEN);
			    bc.pageReconstruction2();
			}
			
		    }
		    
		}else{
		    int left_page= bc.getLeftPageNum();
		    int right_page= bc.getRightPageNum();

		    if( left_page == config.lastPage && scene.selected_name == config.TURN_RIGHT ){
			bc.setSelectedBackCover();
			double secTurningInterval= config.getTurningInterval(); 
			double secs= 0;
			double t;
			
			while(secs < numSecsToRun ){
			    secs += secTurningInterval;
			    t= secs/numSecsToRun;
			    bc.stepBackCoverState(t);
			    
			    try { sleep(turningInterval ); }       // in  msecs
			    catch (InterruptedException e) {};
			}
			
			//turning termination
			bc.setCoverState(CLOSE);
			bc.pageReconstruction2();
			
		    }else if( right_page == config.firstPage && scene.selected_name == config.TURN_LEFT){
		       //close the front cover
			//turning the front cover
			bc.setSelectedFrontCover();
			//bc.jump_SelectedOneStaticPage();
			//bc.jump_theOtherStaticPage();
			
			//double numSecsToRun= 3;
			double secTurningInterval= config.getTurningInterval(); 
			double secs= 0;
			double t;

			while(secs < numSecsToRun ){
			    secs += secTurningInterval;
			    t= secs/numSecsToRun;
			    bc.stepFrontCoverState(t);
			    
			    try { sleep(turningInterval ); }       // in  msecs
			    catch (InterruptedException e) {};
			}
			//rotate clockwisely until closed
			
			bc.setCoverState(CLOSE);
			bc.pageReconstruction2();

		    }else{
			//turning internal page
			bc.setSelectedPage3(scene.selected_name);
			config.setRegion(1); //random ???
			bc.setDynamicPage();
			bc.stepPage_State(0);
			Scene.readytodisplayMovingPage= true;
			bc.jump_SelectedOneStaticPage();// update static page's page number and texturename
			
			//scene.selected_name= bc.getSelectedpage();
			double secTurningInterval= config.getTurningInterval(); 
			double secs= 0;
			double t;
			
			while(secs < numSecsToRun ){
			    secs += secTurningInterval;
			    t= secs/numSecsToRun;
			    bc.stepPage_State(t);
			    
			    try { sleep(turningInterval ); }       // in  msecs
			    catch (InterruptedException e) {};
			}
		    
			bc.jump_theOtherStaticPage();
			bc.updatePapers();
			bc.update_StaticPages();
			bc.pageReconstruction2();
			
			//reset
			Scene.readytodisplayMovingPage= false;
			
		    }

		   
		}
*/
		scene.selection_mode= UNSELECTED;
		scene.turning_mode= NOT_ANIMATING;// reset the restriction
		//}else{
		//System.out.println("!!!Left page");

		//}
	}
	
	else{ 
	    //pages has been selected
	    
	    if(Ttp.commentslevel >= 3){
		System.out.println("\tPage has been selected");
	    }
	    
	    /*page*/
	    if(bc.getSelectedSide() == RIGHT){
		if(Ttp.commentslevel >= 3){
		    System.out.println("\tRight page has been selected");
		}
		
		// turning back
		while( !bc.isRightPage_Rest()){
		    if(Ttp.commentslevel >= 3){
			System.out.println("angleStep: " + angleStep);
		    }
		    try {
			sleep((int)(msecsDisplayInterval));
			//suspend 
			synchronized(this) {
			    while (threadSuspended){
				if(Ttp.commentslevel >= 3){
				    System.out.println("Animation.java: run() getSuspended" );
				}
				wait();
			    }
			}
		    }catch (InterruptedException e) { }
		
		    bc.rotateRightPage(angleStep);
		    
		}
	
		if(bc.getPageTurningAngle() < config.HALF_WAY){
		    bc.step_back_step_SelectedOneStaticPage();
		    callback.actionPerformed(new ActionEvent(this, 0, "PAGE_UNSELECTED_RESTORE")); //PAGE_UNSELECTED_RESTORE
		}else{
		    bc.jump_theOtherStaticPage();
		    callback.actionPerformed(new ActionEvent(this, 0, "PAGE_UNSELECTED_UPDATE")); //PAGE_UNSELECTED_UPDATE
		    bc.updatePapers();
		    bc.update_StaticPages();
		    bc.pageReconstruction2();
		}
		
	    }else{// left page has been selected
		if(Ttp.commentslevel >= 3){
		    System.out.println("\tLeft page has been selected");
		    //System.out.println("Page rotation angle: " + bc.getPageTurningAngle());
		}
		
				
		// turning back
		while( !bc.isLeftPage_Rest()){
		    if(Ttp.commentslevel >= 3){
			System.out.println("angleStep: " + angleStep);
		    }
		    
		    try {
			sleep((int)(msecsDisplayInterval));
			//suspend 
			synchronized(this) {
			    while (threadSuspended){
				if(Ttp.commentslevel >= 3){
				    System.out.println("Animation.java: run() getSuspended" );
				}
				wait();
			    }
			}
		    }catch (InterruptedException e) { }
		    
		    bc.rotateLeftPage(angleStep);
		}
		    
		//animation termination
		if(bc.getPageTurningAngle() < -config.HALF_WAY){
		    bc.jump_theOtherStaticPage();
		    callback.actionPerformed(new ActionEvent(this, 0, "PAGE_UNSELECTED_UPDATE")); //PAGE_UNSELECTED_UPDATE
		    bc.updatePapers();
		    bc.update_StaticPages();
		    bc.pageReconstruction2();
		}else{
		    bc.step_back_step_SelectedOneStaticPage();
		    callback.actionPerformed(new ActionEvent(this, 0, "PAGE_UNSELECTED_RESTORE")); //PAGE_UNSELECTED_RESTORE
		}
		
	    }

	    if(Ttp.commentslevel >= 3){
		System.out.println("\t Animation.java: run() selection_mode == UNSELECTED");
	    }
	    //if texture still rendering
	    scene.selection_mode= UNSELECTED;
	    scene.turning_mode= NOT_ANIMATING;// reset the restriction
	    scene.setCursor(cursorManager.getCursor(IDLE));
	    
	}//end of page
	//scene.repaint();
	if(Ttp.commentslevel >= 2){
	    System.out.println("Animation.java: Finishing running");
	}
	
	
    }//end of run

   
    public void kill() 
    {
        // Called to terminate the separate execution thread of this controller
        killed = true;
        interrupt();
    }

    public synchronized void setSuspend()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("\tAnimation.java: setSuspend()");
	}
	this.threadSuspended = true;
	scene.turning_mode= NOT_ANIMATING;// reset the restriction
    }

    public synchronized void setAwake()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("\tAnimation.java: setAwake()");
	}

	this.threadSuspended = false;
	scene.turning_mode= ANIMATING;// reset the restriction
	notifyAll();
    }

    public boolean iskilled()
    {
	return killed;
    }

    public void setStop()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("\tAnimation.java: setStop()");
	}
	stop= true;
	interrupt();
    }

    public void swapDirection()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("\tAnimation.java: swapDirection()");
	}
	swap= true;
	interrupt();
    }
    
    public void swapDirection(int delta_x)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("\tAnimation.java: swapDirection() with delta_x");
	}
	swap= true;
	this.delta_x= delta_x;
	interrupt();
    }
    
    
    
}
