package TTPproject;

import gl4java.awt.GLCanvas;
import gl4java.utils.textures.*;
import gl4java.GLFunc;
import gl4java.GLEnum;

import javax.vecmath.*;


class BezierCurve 
    implements CONSTANTS, GLEnum{
    
    Point3d []ctrl_points= null;
    Point3d[][] pts= null;

    
    Vector3d [] jackVector= null;
    double [] length;
    
    int n= 0; // number of points
    double unit_length= 0;
    byte side;

    float curve_thickness= 1.f;

    public BezierCurve(int n, float thickness)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierCurve.java: BezierCurve()");
	}

	int i;
	this.curve_thickness= thickness;

	ctrl_points= new Point3d[4];
	
	for(i=0; i<4; i++){
	    ctrl_points[i]=new Point3d(0,0,0);
	}
	
	this.n= n;
	pts= new Point3d[2][n];//'0' for unmodified, and '1' is for after refitted
	jackVector= new Vector3d[n];
	for(i=0; i<n; i++){
	    pts[0][i]=new Point3d(0,0,0);
	    pts[1][i]=new Point3d(0,0,0);
	    jackVector[i]= new Vector3d(0,0,0);
	}
	
	length= new double[n-1];
	
    }



    public void setKnots(int index, Point3d knot)
    {
	ctrl_points[index].set(knot);
    }

    public void setKnots(Point3d [] knots)
    {
	//System.arraycopy(knots, 0, ctrl_points, 0, knots.length);
	Utilities.arrayCopy(knots, 0, ctrl_points, 0, knots.length);
    }
   
    public void setAll(double unit_len, byte side)// apart from knots
    {
	set_unitLen(unit_len);
	set_side(side);
    }
    
    public void set_unitLen(double unitlength)
    {
	unit_length= unitlength;
    }

    public void set_side(byte side)
    {
	this.side= side;
    }
    

    public void genCurves()
    {
	//System.out.println("BezierCurve: genCurves");
	int i;
	double p;

	for(i=0; i<n; i++){
	    p= (double)i/(double)(n-1.0);
	    pts[0][i].set(bernstein(p, ctrl_points));
	}
	
    }
    
    public Point3d [] getPts( int index)
    {
	return pts[index];
    }

    public Vector3d [] getVectors()
    {
	return jackVector;
    }

    /*
      get the first point 
     */
    public Point3d getPoint()
    {
	return pts[1][0];

    }

    
    void refitting()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierCurve.java: refitting()");
	}
	
	CurveManipulator.calculate_curve_length_1(pts[0],
						  length);
	//System.out.println("BezierCurve: lengthcal");
	CurveManipulator.refitting(n, unit_length, 
				   pts[0], pts[1],
				   length );
	//System.out.println("BezierCurve: refitting");
	//CurveManipulator.calculate_curve_length_1(pts[1], length);
    }

    void genJackVector()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierCurve.java: genJackVector()");
	}

	int i;
	double disx;
	Point3d ref= new Point3d();
	for(i=0; i<n; i++){
	    if(side == LEFT)
		disx= -(i*unit_length);
	    else
		disx= i*unit_length;

	    ref.set( disx + pts[1][0].x,
		     pts[1][0].y,
		     pts[1][0].z );
	    jackVector[i]= new Vector3d(pts[1][i].minus(ref)); 
	}
	
    }

    public void evaluate()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierCurve.java: evaluate()");
	}
	
	genCurves();
	refitting();
	genJackVector();
    }

    /*
      without genJackVector()
     */
    public void evaluate2()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierCurve.java: evaluate2()");
	}

	genCurves();
	refitting();
    }

    
    void evaluate3(Point3d [] another_pts, int index )
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierCurve.java: evaluate3()");
	}
	
	Utilities.arrayCopy(another_pts, 0, pts[index], 0, another_pts.length);
	genJackVector();
    }

    //debug
    void displayPts(GLCanvas canvas, int  index, Color3f color)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierCurve.java: displayPts()");
	}

	Point3d ref= new Point3d();
	Point3d endpts= new Point3d();
	GLFunc gl = canvas.gl;
	gl.glPointSize(5.0f);
	int i;
	double disx;
	
	gl.glPushAttrib(gl.GL_ALL_ATTRIB_BITS);
	
	gl.glColor3f(color.x, color.y, color.z);
	gl.glBegin(canvas.GL_POINTS);
	for(i=0; i< n; i++){
	    gl.glVertex3dv( pts[index][i].get3dv());
	}
	gl.glEnd();
	
	gl.glPopAttrib();

    }

   
    //debug
    public void displayCurve(GLCanvas canvas, int index, float [] c)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierCurve.java: displayCurve()");
	}

	GLFunc gl = canvas.gl;
	int i;

	gl.glPushAttrib(gl.GL_ALL_ATTRIB_BITS);
	gl.glLineWidth(curve_thickness);
	gl.glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, c);	
	/*gl.glColor3f(color.x,
		     color.y,
		     color.z );*/
	gl.glBegin(GL_LINE_STRIP);
	for(i=0; i< n-1; i++){
	    gl.glVertex3dv(pts[index][i].get3dv());
	    gl.glVertex3dv(pts[index][i+1].get3dv());
	}
	gl.glEnd();
	gl.glPopAttrib();
    }

       
    Point3d bernstein(double u, Point3d [] p) 
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierCurve.java: berstein()");
	}
	
	Point3d	a, b, c, d, r;
	
	a= p[0].mul( Math.pow( 1-u,3.0 ) ); // P0*(1-u)^3 
	b= p[1].mul( 3.0 * u* Math.pow( 1-u , 2.0) );//P1*3*u*(1-u)^2 
	c= p[2].mul( 3.0 * Math.pow(u,2.0) * (1-u));//P2*3*u^2*(1-u)
	d= p[3].mul( Math.pow(u,3.0) ); //P3* u^3
	r= a.plus(b).plus(c).plus(d);
	
	return r;
	
    }

    Point3d [] points(int index)
    {
	return pts[1];

    }



}
