package TTPproject;

import gl4java.awt.GLCanvas;
import gl4java.utils.textures.*;
import gl4java.GLFunc;
import gl4java.GLEnum;

import javax.vecmath.*;


class BezierPatch 
    implements CONSTANTS, GLEnum{
    
    int nx= 0, nz=0;
    Point3d [][] ctrl_points= null;
    Point3d [][] resultpts= null;

    public BezierPatch(int nx, 
		       int nz 
		       
		       )
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierPatch.java: BezierPatch()");
	}
	
	this.nx= nx;
	this.nz= nz;
	ctrl_points= new Point3d[4][4];
	resultpts= new Point3d[nz][nx];
	
	int r, c;
	for(r=0; r<4; r++){
	    for(c=0; c<4; c++){
		ctrl_points[r][c]=new Point3d();
	    }
	}

	
	for(r=0; r<nz; r++){
	    for(c=0; c<nx; c++){
		resultpts[r][c]=new Point3d();
	    }
	}
	
    }

    public void setControlPoints(Point3d [][] ctrl_pts)
    {
	int r, c;
	for(r=0; r<4; r++){
	    for(c=0; c<4; c++){
		this.ctrl_points[r][c].set(ctrl_pts[r][c]);
	    }
	}
	
	
    }


    /*
      for( u = 0; u < 10; u++ )
      {
         blend row 0 control points -> new control point
         blend row 1 control points -> new control point
         blend row 2 control points -> new control point
         blend row 3 control points -> new control point
      
	for( v = 0; v < 10; v++ )
        {
          blend 4 new control points -> point on surface
	}
      }


    */
    public void genPatch()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierPatch.java: evaluate()");
	}
	
	Point3d [] newCtrlPoints= new Point3d[4];
	double p, q;
	int r, c;

	for(r= 0; r<nz; r++){
	    q= (double)r/(double)(nz-1.0);
	    newCtrlPoints[0]= bernstein(q, ctrl_points[0] );
	    newCtrlPoints[1]= bernstein(q, ctrl_points[1] );
	    newCtrlPoints[2]= bernstein(q, ctrl_points[2] );
	    newCtrlPoints[3]= bernstein(q, ctrl_points[3] );
	    
	    for(c=0; c<nx; c++){
		p= (double)c/(double)(nx-1.0);
		resultpts[r][c].set( bernstein(p, newCtrlPoints ));
	    }
	}
    }
    
    Point3d bernstein(double u, Point3d [] p) 
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("BezierCurve.java: berstein()");
	}
	
	Point3d	a, b, c, d, r;
	
	a= p[0].mul( Math.pow( 1-u,3.0 ) ); // P0*(1-u)^3 
	b= p[1].mul( 3.0 * u* Math.pow( 1-u , 2.0) );//P1*3*u*(1-u)^2 
	c= p[2].mul( 3.0 * Math.pow(u,2.0) * (1-u));//P2*3*u^2*(1-u)
	d= p[3].mul( Math.pow(u,3.0) ); //P3* u^3
	r= a.plus(b).plus(c).plus(d);
	
	return r;
	
    }

    public void evaluate(GLCanvas canvas, int textureName)
    {
	GLFunc gl = canvas.gl;

	gl.glPushAttrib(GL_ALL_ATTRIB_BITS); 
	float [] white= {1.f, 1.f, 1.f, 1.f};

	gl.glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, white);
	gl.glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, white);
	gl.glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 1);

	int r,c;
	for (r = 0; r < nz-1 ; r++) {
	    gl.glBegin(GL_QUAD_STRIP);
            for (c = 0; c < nx; c++) {
		//		gl.glMultiTexCoord2dARB(GL_TEXTURE0_ARB, (c / nx_intervals), -(r+1) / nz_intervals );
		//gl.glNormal3dv(normals[i+1][j].get3dv());
		gl.glVertex3dv(resultpts[r+1][c].get3dv());
		
		
		//gl.glMultiTexCoord2dARB(GL_TEXTURE0_ARB, (c / nx_intervals), -r / nz_intervals);
		//gl.glNormal3dv(normals[i][j].get3dv());
		gl.glVertex3dv(resultpts[r][c].get3dv());
	    }
            
	    gl.glEnd();
        
	}

	/*gl.glBegin (GL_POINTS);
	
	int r,c;
	for(r= 0; r < nz-4; r++){
	    for(c= 0; c < nx; c++){
		gl.glVertex3dv(resultpts[r][c].get3dv());
	    }
	}
	
	gl.glEnd();*/
		

	gl.glPopAttrib();

    }

    
    public void evaluate(GLCanvas canvas,  byte command,  int textureName, int materialName) //changed
    {	
	if(Ttp.commentslevel >= 2){
	    System.out.println("DynamicPage.java: displayFirstSurface()" );
	}
	
	
	GLFunc gl = canvas.gl;
	
	int i, j;
	double nx_intervals= nx - 1.0;
	double nz_intervals= nz - 1.0;
	boolean ef= false, et= false;
	
	switch(command){
	case LINE:{ 
	    if(Ttp.commentslevel >= 3){
		System.out.println("line");
	    }
	    ef= false; et=false;
	    break;}
	case FILL:{
	    if(Ttp.commentslevel >= 3){
		System.out.println("fill" );
	    }
	    ef= true; et=false;
	    break;}
	case TEXTURE:{ 
	    if(Ttp.commentslevel >= 3){
		System.out.println("texture" );
	    }
	    ef= true; et=true;
	    break;}
	}

	gl.glPushAttrib(GL_ALL_ATTRIB_BITS); 
	float [] white= {1.f, 1.f, 1.f, 1.f};

	gl.glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, white);
	gl.glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, white);
	gl.glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 1);
	    
	if(ef)
	    gl.glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	else
	    gl.glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	if(et){
	    //gl.glEnable(GL_TEXTURE_2D);
	    //gl.glBindTexture(GL_TEXTURE_2D, textureName_up);
	    gl.glActiveTextureARB(GL_TEXTURE0_ARB);
	    gl.glEnable(GL_TEXTURE_2D);
	    gl.glBindTexture(GL_TEXTURE_2D, textureName);
	    
	    gl.glActiveTextureARB(GL_TEXTURE1_ARB);//changed
	    gl.glEnable(GL_TEXTURE_2D);
	    gl.glBindTexture(GL_TEXTURE_2D, materialName);

	}
	else 
	    gl.glDisable(GL_TEXTURE_2D);

	int r,c;
	double u, v;
	for (r = 0; r < nz - 1; r++) {
	    gl.glBegin(GL_QUAD_STRIP);
            for (c = 0; c < nx; c++) {
		//gl.glMultiTexCoord2dARB(GL_TEXTURE0_ARB, (c / nx_intervals), (r+1) / nz_intervals );
		
		u= 1.0 - c / nx_intervals;
		v= (r+1) / nz_intervals;
				
		gl.glMultiTexCoord2dARB(GL_TEXTURE0_ARB, v, u);
		gl.glMultiTexCoord2dARB(GL_TEXTURE1_ARB, v, u); //changed
		//gl.glMultiTexCoord2dARB(GL_TEXTURE0_ARB,  c / nx_intervals, (r+1) / nz_intervals );
		//gl.glNormal3dv(normals[i+1][j].get3dv());
		gl.glVertex3dv(resultpts[r+1][c].get3dv());

		u= 1.0 - c / nx_intervals;
		v= r / nz_intervals;
					
		gl.glMultiTexCoord2dARB(GL_TEXTURE0_ARB, v, u);
		gl.glMultiTexCoord2dARB(GL_TEXTURE1_ARB, v, u); //changed
		//gl.glMultiTexCoord2dARB(GL_TEXTURE0_ARB, c / nx_intervals, r / nz_intervals);
		//gl.glNormal3dv(normals[i][j].get3dv());
		gl.glVertex3dv(resultpts[r][c].get3dv());
	    }
	    gl.glEnd();
            
        }

	gl.glPopAttrib();
    }



}
