/*
  Zoom-in based on distance between eye and lookAt
*/

package TTPproject;

import javax.vecmath.Point3d;
import javax.vecmath.Point2d;
import javax.vecmath.Color4f;
import javax.vecmath.Vector3d;
import java.io.File;
import java.io.BufferedReader;
import java.io.StreamTokenizer;
import java.io.IOException;
import java.io.Reader;
import java.io.FileReader;

import java.awt.Rectangle;

public class Configuration 
    implements CONSTANTS{
    
    Parameters parameters= null;
    
    //doing anchoring or not
    int running_type;
    int anchoring_type;
    int interval; // must be even number
    String anchoring_file; //anchoring file name
    String page_log_file; //anchoring file name
    
    //window 
    Rectangle windowbounds= null; //initial window size
    //viewport 
    Rectangle viewportbounds= null; //initial viewport size


    //page content size
    int imageWidth_highresol= 0;
    int imageHeight_highresol= 0;
    
    int imageWidth_lowresol= 0;
    int imageHeight_lowresol= 0;
    
    int maxlevel= 1;
    int imageDelta= 64;
    int imageMultiple= 2;
    int increMethod= ADD;

    //trackball zooming param
    double lower_zoom_factor= 1.15f;
    double upper_zoom_factor= 1.5f;
    float deltaZ= .0926f;
    //new zoom-in parameters
    public double lengthfactor= 1.f;
    public double lower_length_factor= 0.58f;
    public double upper_length_factor= 2.61f;
        
    
    double fixed_distance_between_viewingplane_zCentre= 0;
    double distance_between_viewingplane_zCentre= 0;
    double distance_betweenEye_LookAt= 0; //fixed value
    Point3d neweye= null;
        
    // cover & page param
    int nx; // n points along x-axis
    int nz; // n points along z-axis
    double elementLength; // unit length
    double cover_margin; // margin between page and cover
    double cover_width;
    
    double cover_thickness;
    double cover_ratio;
    double spinef2_height;
    double spinef5_height;
    double book_thickness; 
    
    int pdf_total_pages;
    int total_content_pages; // excluding from first and second pages
    int thefirstPdfpage= 3;
    int thefirstPNGpage= 2;
    int total_pages;//deprecated
    int total_papers;

    int firstPage, lastPage; 
    
    double paperWidth;
    double paperHeight;
    double paperThickness;

    int current_paper; //testing
    int startingPage;

    double rotAround_ck2; //testing
    int init_coverState; //testing

    double zMax; // cover height
    double xMax; // cover width


    //viewing param
    double EYE_Y_FACTOR, EYE_Z_FACTOR;
    double FOV, aspect, near, far; // perspective param
    double initial_FOV, initial_aspect, initial_near, initial_far;
    
    double centerZ;
    Point3d eye= null;
    Point3d lookAt= null;
    Point3d initial_eye= null;
    Point3d initial_lookAt= null;

    //reading mode param
    double howmuchtoturn= 0;
    double readingplane_aspect= 1;// fixed
    double readingplane_dvalue= 0;// how far the reading plane is away from the viewpoint
    double alpha= 0;
    double ALPHA_DEGREE= 0;
    double updateInterval= 0;
    int totalSteps= 100;

    Vector3d upVector= null;
    Vector3d aVector= null;
    Vector3d bVector= null;
    

    //extra selection names
    int FRONT_COVER, BACK_COVER, LEFT_EDGE, RIGHT_EDGE; 
    int STOP_BUTTON, FORWARD_BUTTON, BACKWARD_BUTTON, INFO_BUTTON, RESET_BUTTON, TURN_LEFT, TURN_RIGHT;
   

    //particles reader
    ParticlesReader3 PR= null;
    int region;
    
    //page 
    float[] PAGE_COLOUR;
    float[] PAGE_SPECULAR;
    float   PAGE_SHINESS;
    
    //cover
    float[] COVER_COLOUR;
    float[] COVER_SPECULAR;
    float   COVER_SHINESS;

    //light 0
    float[] LIGHT0_POSITION;
    float[] AMBIENT_LIGHT_COLOUR ;
    float[] DIFFUSE_LIGHT_COLOUR ;
    float[] SPECULAR_LIGHT_COLOUR;
    float[] lmodel_ambient;
  
    //dropping
    double timeStep;
    double displayInterval;
    double angleStep;

    //flipping
    double flippingAngleStep;
    double pausing;
    double turningInterval;

    Point2d [] points= null;

    int bufferingOption= 0;
    int imageformat= 0;
    
    // Reader
    HtmlConverter converter= null;
    public int num_dynamic_tex_obj; 
    public int threasholdK; //could be called midpoint

    //animation
    int HALF_WAY; //in degree
    double SLOW; 
    double FAST;

    double SHORT = .5; // in second
    double LONG = 3; // in second

    //duration
    long LOWER_GESTURE_TIME;
    long UPPER_GESTURE_TIME;
    //displacement
    int LOWER_DISPLACEMENT;
    int UPPER_DISPLACEMENT; //x-axis displacement pixels 
    
    //displaying anchor pages
    boolean displayAnchorpoint= true;
    //snapping
    boolean snapping= true;
    boolean progressiveUpate= false;
    //false means i am in 3d mode
    //true means i am in reading mode
    
    
    /**
     * Constructor 
     * @param filename
     */
    public Configuration(Parameters parameters){
	
	if(Ttp.commentslevel >= 2){
	    System.out.println("Configuaration.java: Configuaration()");
	}

	this.parameters= parameters;
	windowbounds= parameters.getWindowBounds();
	viewportbounds= parameters.getViewportBounds();
    }

    

    public void doInitialized()
    {

	double temp;
		
	if(Ttp.commentslevel >= 2){
	    System.out.println("Configuaration.java: doInitialized()");
	}
	
	//particles
	PR= new ParticlesReader3(Ttp.prefix);// read in all at once
	
	//pdf file reader
	/*
	  Read in pdf file
	  Extract the name of the file
	  Extract the number of frames (total pages)
	  Performing conversion
	    for each page
              re-scaling
	      possibly image enhancement
	      relabel it
	*/
	
	
	if(PR.isOK()){
	    //paper size
	    Ttp.messagesState= 1;
	    nx= PR.getNx();
	    nz= PR.getNz();
	    elementLength= PR.getMeshElementLength();
	    paperWidth= (nx-1) * elementLength;
	    paperHeight= (nz-1) * elementLength;
	    if(Ttp.commentslevel >= 3){
		System.out.println("paper width: " + paperWidth);
		System.out.println("paper length: " + paperHeight);
	    }
		
	    //new 9_6
	    try{
		String fname= Ttp.prefix + parameters.getConfigfile();
		
		File testData= new File(fname);
		Reader r = new BufferedReader(new FileReader(testData));
		StreamTokenizer st = new StreamTokenizer(r);
		
		st.wordChars('_', '_'); //Specifies that all characters c in the range low <= c <= high  are word constituents.
		int tokentype;
		int EQUAL= 0;
		String flag;

		if(Ttp.commentslevel >= 3){
		    System.out.println("Reading from the Configuration file ");
		}
		
		for(;;){
		    tokentype= st.nextToken();
		    if(tokentype == StreamTokenizer.TT_EOF) break;
		    //System.out.println("Not finished yet");
		    flag= st.sval.toUpperCase(); 
		    if(Ttp.commentslevel >= 3){
			System.out.println(flag);
		    }
		    st.nextToken();
		    
		    if(flag.compareTo("PT") == EQUAL){
			paperThickness= st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("paper thickness: " + paperThickness );
			}
			
		    }else if(flag.compareTo("CM") == EQUAL){
			cover_margin= st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("Cover margin: " + cover_margin );
			}
			
		    }else if(flag.compareTo("CT") == EQUAL){
			temp= st.nval; 
			
		    }else if(flag.compareTo("CRATIO") == EQUAL){
			cover_ratio= st.nval; 
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("cover ratio: " + cover_ratio);
			}
			
		    }else if(flag.compareTo("SPINEFACETWOHEIGHT") == EQUAL){
			temp= st.nval;// ??? 
			
		    }else if(flag.compareTo("SPINEFACEFIVEHEIGHT") == EQUAL){
			temp= st.nval;// ??? 
			//System.out.println("spine face five height: " + spinef5_height);
		    }else if(flag.compareTo("CP") == EQUAL){ // testing
			startingPage= (int)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("starting page: " + startingPage);
			}
			
		    }else if(flag.compareTo("ROTCK2") == EQUAL){ // testing
			rotAround_ck2= st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("Rotate around c2: " + rotAround_ck2);
			}
			
		    }else if(flag.compareTo("COVERSTATE") == EQUAL){ // testing
			init_coverState= (int)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("init_coverState: " + init_coverState);
			}
			
			if(init_coverState == 0){
			    init_coverState=CLOSE;
			   
			    if(Ttp.commentslevel >= 3){
				System.out.println("init_coverState=CLOSE");
			    }
			}
			else{
			    init_coverState=OPEN;
			    
			    if(Ttp.commentslevel >= 3){
				System.out.println("init_coverState=OPEN");
			    }
			}

		    }else if(flag.compareTo("EYEYFACTOR") == EQUAL){
			EYE_Y_FACTOR= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Eye y factor: "+ EYE_Y_FACTOR);
			}
			
		    }else if(flag.compareTo("EYEZFACTOR") == EQUAL){
			EYE_Z_FACTOR= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Eye z factor: " + EYE_Z_FACTOR);
			}
		    }
		    else if(flag.compareTo("ALPHA") == EQUAL){
			ALPHA_DEGREE= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("alpha in degree: " + ALPHA_DEGREE);
			}
		    }
		    
		    else if(flag.compareTo("PAGESPECULAR") == EQUAL){
			PAGE_SPECULAR= new float[4];
			
			PAGE_SPECULAR[0]= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("page specular[0]= " + PAGE_SPECULAR[0]);
			}
			
			st.nextToken();
			PAGE_SPECULAR[1]= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("page specular[1]= " + PAGE_SPECULAR[1]);
			}
			
			st.nextToken();
			PAGE_SPECULAR[2]= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("page specular[2] " + PAGE_SPECULAR[2]);
			}
			
			st.nextToken();
			PAGE_SPECULAR[3]= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("page specular[3] " + PAGE_SPECULAR[3]);
			}
						
		    }else if(flag.compareTo("PAGESHINESS") == EQUAL){
			PAGE_SHINESS= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("page shiness " + PAGE_SHINESS);
			}
			
		    }else if(flag.compareTo("COVERCOLOR") == EQUAL){
			COVER_COLOUR= new float[4];
			
			COVER_COLOUR[0]= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("cover color[0]= " + COVER_COLOUR[0]);
			}
			
			st.nextToken();
			COVER_COLOUR[1]= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("cover color[1]= " + COVER_COLOUR[1]);
			}
			
			st.nextToken();
			COVER_COLOUR[2]= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("cover color[2] " + COVER_COLOUR[2]);
			}
			
			st.nextToken();
			COVER_COLOUR[3]= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("cover color[3] " + COVER_COLOUR[3]);
			}

		    }else if(flag.compareTo("COVERSPECULAR") == EQUAL){
			COVER_SPECULAR= new float[4];
			
			COVER_SPECULAR[0]= (float)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("cover specular[0]= " + COVER_SPECULAR[0]);
			}
			
			st.nextToken();
			COVER_SPECULAR[1]= (float)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("cover specular[1]= " + COVER_SPECULAR[1]);
			}
			
			st.nextToken();
			COVER_SPECULAR[2]= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("cover specular[2] " + COVER_SPECULAR[2]);
			}
			
			st.nextToken();
			COVER_SPECULAR[3]= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("cover specular[3] " + COVER_SPECULAR[3]);
			}
			
		    }else if(flag.compareTo("COVERSHINESS") == EQUAL){
			COVER_SHINESS= (float)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("cover shiness " + COVER_SHINESS);
			}
			
		    }else if(flag.compareTo("LMODELAMBIENT") == EQUAL){
			lmodel_ambient= new float[4];

			lmodel_ambient[0]= (float)st.nval;st.nextToken();
			lmodel_ambient[1]= (float)st.nval;st.nextToken();
			lmodel_ambient[2]= (float)st.nval;st.nextToken();
			lmodel_ambient[3]= (float)st.nval;
						
		    }else if(flag.compareTo("LIGHT0POSITION") == EQUAL){
			LIGHT0_POSITION= new float[4];

			LIGHT0_POSITION[0]= (float)st.nval;st.nextToken();
			LIGHT0_POSITION[1]= (float)st.nval;st.nextToken();
			LIGHT0_POSITION[2]= (float)st.nval;st.nextToken();
			LIGHT0_POSITION[3]= (float)st.nval;
						
		    }else if(flag.compareTo("AMBIENTLIGHTCOLOR") == EQUAL){
			AMBIENT_LIGHT_COLOUR= new float[4];
			
			AMBIENT_LIGHT_COLOUR[0]= (float)st.nval;st.nextToken();
			AMBIENT_LIGHT_COLOUR[1]= (float)st.nval;st.nextToken();
			AMBIENT_LIGHT_COLOUR[2]= (float)st.nval;st.nextToken();
			AMBIENT_LIGHT_COLOUR[3]= (float)st.nval;
						
		    }else if(flag.compareTo("DIFFUSELIGHTCOLOR") == EQUAL){
			DIFFUSE_LIGHT_COLOUR= new float[4];
			
			DIFFUSE_LIGHT_COLOUR[0]= (float)st.nval;st.nextToken();
			DIFFUSE_LIGHT_COLOUR[1]= (float)st.nval;st.nextToken();
			DIFFUSE_LIGHT_COLOUR[2]= (float)st.nval;st.nextToken();
			DIFFUSE_LIGHT_COLOUR[3]= (float)st.nval;
						
		    }else if(flag.compareTo("SPECULARLIGHTCOLOR") == EQUAL){
			SPECULAR_LIGHT_COLOUR= new float[4];
			
			SPECULAR_LIGHT_COLOUR[0]= (float)st.nval;st.nextToken();
			SPECULAR_LIGHT_COLOUR[1]= (float)st.nval;st.nextToken();
			SPECULAR_LIGHT_COLOUR[2]= (float)st.nval;st.nextToken();
			SPECULAR_LIGHT_COLOUR[3]= (float)st.nval;
						
		    }else if(flag.compareTo("POINTS") == EQUAL){ // unsure what it is for
			points= new Point2d[8];
			
			for(int i= 0; i< 4; i++){
			    points[i]= new Point2d(0,0); 
			}

			points[0].x= (float)st.nval; st.nextToken();
			points[0].y= (float)st.nval; st.nextToken();
			points[1].x= (float)st.nval; st.nextToken();
			points[1].y= (float)st.nval; st.nextToken();
			points[2].x= (float)st.nval; st.nextToken();
			points[2].y= (float)st.nval; st.nextToken();
			points[3].x= (float)st.nval; st.nextToken();
			points[3].y= (float)st.nval; 			
			
		    }else if(flag.compareTo("DT") == EQUAL){
			timeStep= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("time step: " +  timeStep );
			}
			
		    }else if(flag.compareTo("TDISPLAY") == EQUAL){
			displayInterval= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("display interval: " + displayInterval );
			}
			
		    }else if(flag.compareTo("ANGLESTEP") == EQUAL){
			angleStep= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("angle step: " + angleStep );
			}
			
		    }else if(flag.compareTo("FLIPPING_ANGLESTEP") == EQUAL){
			flippingAngleStep= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Flipping angle step: " + flippingAngleStep );
			}
			
		    }else if(flag.compareTo("PAUSING") == EQUAL){
			pausing= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Pausing: " + pausing );
			}
			
		    }else if(flag.compareTo("TURNING_INTERVAL") == EQUAL){
			turningInterval= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Turning interval: " + turningInterval );
			}
			
		    }else if(flag.compareTo("TEXTURENAMESIZE") == EQUAL){
			num_dynamic_tex_obj= (int)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("texture name size: " +  num_dynamic_tex_obj);
			}
			
		    }else if(flag.compareTo("UPDATE_INTERVAL") == EQUAL){
			updateInterval= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Update interval: " + updateInterval );
			}
		    }else if(flag.compareTo("TOTALSTEPS") == EQUAL){
			totalSteps= (int)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Total Steps: " + totalSteps );
			}
		    }

		    /*
		    else if(flag.compareTo("THREASHOLDK") == EQUAL){
			threasholdK= (int)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("threshold value k: " + threasholdK);
			}
			
		    }else if(flag.compareTo("WINWIDTH") == EQUAL){
			winWidth= (int)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("window width: " + winWidth);
			}
			
		    }else if(flag.compareTo("WINHEIGHT") == EQUAL){
			winHeight= (int)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("window height: " + winHeight);
			}
			
		    }

		    
		    else if(flag.compareTo("STARTX") == EQUAL){
			startx= (int)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("startx: " + startx);
			}
		    }

		    else if(flag.compareTo("STARTY") == EQUAL){
			starty= (int)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("starty: " + starty);
			}
		    }

		    else if(flag.compareTo("VIEWPORT_WIDTH") == EQUAL){
			viewportWidth= (int)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("viewport width: " + viewportWidth);
			}
			
		    }else if(flag.compareTo("VIEWPORT_HEIGHT") == EQUAL){
			viewportHeight= (int)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("viewport height: " + viewportHeight);
			}
			
			}
		    */

		    else if(flag.compareTo("LOW_IMAGEWIDTH") == EQUAL){
			imageWidth_lowresol= (int)st.nval;

			if(Ttp.commentslevel >= 3){
			    System.out.println("Low resolution image width: " + imageWidth_lowresol);
			}
		    }else if(flag.compareTo("LOW_IMAGEHEIGHT") == EQUAL){
			imageHeight_lowresol= (int)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("Low resolution image height: " + imageHeight_lowresol);
			}
			
		    }else if(flag.compareTo("HIGH_IMAGEWIDTH") == EQUAL){
			imageWidth_highresol= (int)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("High resolution image width: " + imageWidth_highresol);
			}
			
		    }else if(flag.compareTo("HIGH_IMAGEHEIGHT") == EQUAL){
			imageHeight_highresol= (int)st.nval;
			
			if(Ttp.commentslevel >= 3){
			    System.out.println("High resolution image height: " + imageHeight_highresol);
			}
			
		    }else if(flag.compareTo("BUFFERINGOPTION") == EQUAL){
			//to do with windowing technique 
			bufferingOption= (int)st.nval;
		    }else if(flag.compareTo("IMAGEFORMAT") == EQUAL){
			
			String image= st.sval.toUpperCase();
			if(Ttp.commentslevel >= 3){
			    System.out.println("Image format"); 
			}
			
			if(image.compareTo("PNG") == EQUAL){
			    if(Ttp.commentslevel >= 3){
				System.out.println("PNG") ;
			    }
			    imageformat= PNG;
			}else if(image.compareTo("JPEG")== EQUAL){
			    if(Ttp.commentslevel >= 3){
				System.out.println("JPEG");
			    } 
			    imageformat= JPEG;
			}else if(image.compareTo("TGA")== EQUAL){
			    if(Ttp.commentslevel >= 3){
				System.out.println("TGA"); 
			    }
			    imageformat= TGA;
			}else{
			    if(Ttp.commentslevel >= 3){
				System.out.println("Unspecified image format"); 
			    }
			}
			
			
		    }else if(flag.compareTo("LEVELS") == EQUAL){
			maxlevel= (int)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Max Levels: " + maxlevel);
			}

		    }else if(flag.compareTo("INCREMENT") == EQUAL){
			String method= st.sval.toUpperCase();
			
			if(Ttp.commentslevel >= 3){
			    System.out.print("Increment method: ");
			}

			if(method.compareTo("ADD") == EQUAL){
			    if(Ttp.commentslevel >= 3){
				System.out.print("add ");
			    }
			    
			    increMethod= ADD;
			    st.nextToken();
			    imageDelta= (int)st.nval;
			    if(Ttp.commentslevel >= 3){
				System.out.println(imageDelta);
			    }
			}else{
			    if(Ttp.commentslevel >= 3){
				System.out.print("multiplied by ");
			    }
			    
			    increMethod= MULTIPLY;
			    st.nextToken();
			    imageMultiple= (int)st.nval;
			    if(Ttp.commentslevel >= 3){
				System.out.println(imageMultiple);
			    }
			}			
			
			
		    }else if(flag.compareTo("LOWER_ZOOM_FACTOR") == EQUAL){
			lower_zoom_factor= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Lower zoom factor: " + lower_zoom_factor);
			}
		    
		    }else if(flag.compareTo("RUNNING_TYPE") == EQUAL){
			if(Ttp.commentslevel >= 3){
			    System.out.println("Running type: " );
			}
			
			String running= st.sval.toUpperCase();// convert to upper case

			if(running.compareTo("PROGRESSIVE") == EQUAL){
			    if(Ttp.commentslevel >= 3){
				System.out.println("Progressive: " );
			    }
			    running_type= NO_ANCHORING;

			}else if(running.compareTo("ANCHOR") == EQUAL){
			    if(Ttp.commentslevel >= 3){
				System.out.println("ANCHOR: "  );
			    }
			    
			    st.nextToken();
			    
			    running_type= ANCHORING;
			    String anchors= st.sval.toUpperCase(); 
			    //deprecated
			    if(anchors.compareTo("STEP") == EQUAL){
				st.nextToken();
				//step
				if(Ttp.commentslevel >= 3){
				    System.out.println("stepping"  );
				}
				anchoring_type= STEPPING;
				interval= (int)st.nval;
				
				if(interval >= 0) // interval must be positive and at least 2
				    interval= 2;
				if(Ttp.commentslevel >= 3){
				    System.out.println("interval= " + interval );
				}
				
			    }
			    
			    //deprecated
			    else if(anchors.compareTo("RANDOM") == EQUAL){
				st.nextToken();
				//random
				if(Ttp.commentslevel >= 3){
				    System.out.println("Random jumping"  );
				}
				anchoring_type= RANDOM_JUMP;
				interval= (int)st.nval;
				
				if(interval <= 0) // interval must be positive and at least 2
				    interval= 2; 
				if(Ttp.commentslevel >= 3){
				    System.out.println("interval= " + interval );
				}
				
			    }
			    
			    else if(anchors.compareTo("FILE") == EQUAL){
				//reading the value from a file
				if(Ttp.commentslevel >= 3){
				    System.out.println("Require Anchoring page from the file"  );
				}
				
				anchoring_type= FROM_FILE;
				//String file= st.sval;
				if(parameters.getfileType() == PDF){
				    anchoring_file= Ttp.pdfprefix + Ttp.ebook_file_name + ".ash";
				}else if(parameters.getfileType() == DJVU){
				    anchoring_file= Ttp.djvusprefix + Ttp.ebook_file_name + ".ash";
				}

				if(Ttp.commentslevel >= 3){
				    System.out.println("Anchoring file: " +  anchoring_file);
				}
			    }
			}
			
		    
		    }else if(flag.compareTo("UPPER_ZOOM_FACTOR") == EQUAL){
			upper_zoom_factor= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Upper zoom factor: " + upper_zoom_factor);
			}
		    }
		    
		    else if(flag.compareTo("HALF_WAY") == EQUAL){
			//dropping angle
			HALF_WAY= (int)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Half way : " + HALF_WAY);
			}
		    
		    }

		    else if(flag.compareTo("SLOW") == EQUAL){
			SLOW= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Slow : " + SLOW);
			}
		    
		    }
		    
		    else if(flag.compareTo("FAST") == EQUAL){
			FAST= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Fast : " + FAST);
			}
		    }
		    
		    else if(flag.compareTo("FAST") == EQUAL){
			FAST= st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Fast : " + FAST);
			}
		    }
		   
		    else if(flag.compareTo("LOWER_GESTURE_TIME") == EQUAL){
			LOWER_GESTURE_TIME= (long)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Lower gesture time : " + LOWER_GESTURE_TIME);
			}
		    }
		    
		    else if(flag.compareTo("UPPER_GESTURE_TIME") == EQUAL){
			UPPER_GESTURE_TIME= (long)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Upper gesture time : " + UPPER_GESTURE_TIME);
			}
		    }

		    else if(flag.compareTo("LOWER_DISPLACEMENT") == EQUAL){
			LOWER_DISPLACEMENT= (int)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Lower displacement : " + LOWER_DISPLACEMENT);
			}
		    }
		    
		    else if(flag.compareTo("UPPER_DISPLACEMENT") == EQUAL){
			UPPER_DISPLACEMENT= (int)st.nval;
			if(Ttp.commentslevel >= 3){
			    System.out.println("Upper displacement : " + UPPER_DISPLACEMENT);
			}
		    }

		    else{
			if(Ttp.commentslevel >= 3){
			    System.out.println("unknown parameters: "+ st.nval);
			}
		    }
		}

		//page color
		PAGE_COLOUR= new float[4];

		//bounds = new Rectangle(startx, starty,  winWidth, winHeight);
		


		float alpha= (float)parameters.getYearLevel()/(float)PERIOD;
		Color4f currentColor= new Color4f(0.0f, 0.0f, 0.0f, 1.0f);
		currentColor.interpolate(BROWN, WHITE, alpha);
		setPageColor(currentColor);
		
		if(parameters.getfileType() == PDF){
		    page_log_file= Ttp.pdfprefix + Ttp.ebook_file_name + ".log";
		}else if(parameters.getfileType() ==  DJVU){
		    page_log_file= Ttp.djvusprefix + Ttp.ebook_file_name + ".log";
		}

		if(Ttp.commentslevel >= 3){
		    System.out.println("Page log file: " +  page_log_file);
		}
		
		
		if(maxlevel > 1){
		    if(increMethod == ADD){ 
			//through addition
			imageWidth_highresol= imageWidth_lowresol + imageDelta * (maxlevel-1);
			imageHeight_highresol= imageHeight_lowresol + imageDelta * (maxlevel-1);
		    }else{
			//through multiplication
			int level= 1;
			imageWidth_highresol= imageWidth_lowresol;
			imageHeight_highresol= imageHeight_lowresol;
			
			while(level < maxlevel){
			    imageWidth_highresol= imageWidth_highresol * imageMultiple;
			    imageHeight_highresol= imageHeight_highresol * imageMultiple;
			    level++;
			}
		    }
		}else{
		    imageWidth_highresol= imageWidth_lowresol;
		    imageHeight_highresol= imageHeight_lowresol;
		}

		//Construct HTMLConverter
		converter= new HtmlConverter(this, //config
					     imageformat, //png, jpeg
					     Ttp.pdfprefix, //pdfs directory
					     Ttp.pagePrefix); // pagecache directory
					     
		
		
		
		pdf_total_pages= parameters.getTotalPages();
		if(pdf_total_pages % 2 == 1){
		    //odd number
		    pdf_total_pages+=1;
		}
		
		//read pdf 
		if(parameters.getfileType() == PDF){
		    if(converter.readFromfile(parameters.getPdffilenames()) == 1){
			if(Ttp.commentslevel >= 3){
			    System.out.println("Unable to read pdf file");
			}
		    }
		    total_content_pages= pdf_total_pages-2; //exclude the first two pages 

		}else if(parameters.getfileType() == DJVU){
		    total_content_pages= pdf_total_pages-4; //exclude the first two and last two pages
		}
		
		
		total_papers= total_content_pages/2;

		firstPage= 0; //page name form //depends on converter
		lastPage= pdf_total_pages - thefirstPdfpage;//page name form 
		
		if(parameters.getfileType() == DJVU){
		    lastPage -= 2; // reserve for backcover and inside backcover
		}
	
		if((lastPage%2) == 0){ //if the last page is Right-hand page
		    lastPage++;
		}
		
		//initialize additional selection names apart from page numbers
		FRONT_COVER= pdf_total_pages; //depends on converter
		BACK_COVER= FRONT_COVER + 1;
		LEFT_EDGE= BACK_COVER + 1;
		RIGHT_EDGE= LEFT_EDGE + 1;
		
		
		STOP_BUTTON= RIGHT_EDGE + 1;  
		FORWARD_BUTTON= STOP_BUTTON + 1; 
		BACKWARD_BUTTON= FORWARD_BUTTON + 1;
		INFO_BUTTON= BACKWARD_BUTTON + 1;
		RESET_BUTTON= INFO_BUTTON + 1;
		TURN_LEFT= RESET_BUTTON + 1;
		TURN_RIGHT= TURN_LEFT + 1;
		
		
		if(Ttp.commentslevel >= 3){
		    System.out.println("pdf_total_pages: " + pdf_total_pages);
		    System.out.println("total_papers: " + total_papers);
		    System.out.println("total_content_pages: " + total_content_pages);
		    System.out.println("firstPage: " + firstPage);
		    System.out.println("****lastPage: " + lastPage);
		    //make sure each one is unique
		    System.out.println("FRONT_COVER: " + FRONT_COVER );
		    System.out.println("BACK_COVER: " + BACK_COVER );
		    System.out.println("LEFT_EDGE: " + LEFT_EDGE );
		    System.out.println("RIGHT_EDGE: " + RIGHT_EDGE );
		    System.out.println("FORWARD_BUTTON: " + FORWARD_BUTTON );
		    System.out.println("STOP_BUTTON: " + STOP_BUTTON );
		    System.out.println("BACKWARD_BUTTON: " + BACKWARD_BUTTON );
		}
		

		cover_width= paperWidth + cover_margin;
		book_thickness= total_papers * paperThickness; 
		double q= (double)pdf_total_pages/(double)MAX_PAGES;
		cover_thickness= MAX_COVER_THICKNESS*q;
		spinef2_height= MAX_SPINEF2_HEIGHT*q;
		spinef5_height= MAX_SPINEF5_HEIGHT*q;
		zMax= paperHeight + cover_margin;
		xMax= paperWidth + cover_margin;
		
		if(Ttp.commentslevel >= 3){
		    System.out.println("cover_width: " + cover_width);
		    System.out.println("book thickness: " + book_thickness);
		    System.out.println("cover thickness: " + cover_thickness);
		    System.out.println("spine face two height: " + spinef2_height);
		    System.out.println("spine face two height: " + spinef5_height);
		    System.out.println("zMax: " + zMax);
		    System.out.println("xMax: " + xMax);
		}
		
		//bufferingOption will be deprecated
		/*switch(bufferingOption){
		case WINDOWING:
		    System.out.println("Windowing technique"); break;
		case DIRECT_ONE:
		    System.out.println("Direct one"); 
		    break;
		case DIRECT_TWO:{
		    System.out.println("Direct two"); 
		    converter.genImages(); //generate series of images
		    break;
		}
		    //case ANCHORING:{
		//System.out.println("Anchoring"); break;
		default:     
		    System.out.println("Unknown buffering option"); 
		    }*/

		//trackball zooming param
		this.deltaZ= calculateDeltaZ(lower_zoom_factor, upper_zoom_factor, maxlevel);
				
		//System.out.println("Finished reading");
		
	    }catch(IOException e){
		if(Ttp.commentslevel >= 3){
		    System.out.println("****IOException error: " + e );
		}
	    }
	    
	    /*used to be*/
	    //initViewingParam();
	    //initZoominParam();
	    //readingplane_aspect= (paperWidth+paperWidth)/paperHeight;
	    //updateReadingPlane_dvalue();
	    /********************/

	    //new order 
	    // fixed values
	    FOV= 45;
	    near= zMax;
	    far= zMax * 30; // may be adjusted according to 3d model
	    centerZ= zMax / 2.0;
	    lookAt= new Point3d(0.0, 0.0, centerZ);
	    upVector= new Vector3d(0.0, 1.0, 0.0);
	    fixed_distance_between_viewingplane_zCentre= 30;
	    //double degree= 63.43494882292201;
	    alpha= Utilities.degree2rad(ALPHA_DEGREE);
	    readingplane_aspect= (paperWidth+paperWidth)/paperHeight; //reading aspect
	    updateAspect(windowbounds.width, windowbounds.height);// window aspect

	    updateReadingPlane_dvalue();
	    // update distance_betweenEye_LookAt
	    distance_betweenEye_LookAt= readingplane_dvalue + fixed_distance_between_viewingplane_zCentre; //adding some constant
	    // update eye position
	    double z, y;
	    z= distance_betweenEye_LookAt * Math.cos(alpha);
	    y= distance_betweenEye_LookAt * Math.sin(alpha);
	    eye= new Point3d(0.0, y, z);
	    //initial_eye= new Point3d(eye);
	    //initial_lookAt= new Point3d(lookAt);
	    
	    
	}else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("Fail to read in particles' positions");
	    }
	}
    }

    public void reset()
    {
	lengthfactor= 1.f;
	snapping= true;
	progressiveUpate= false;
	displayAnchorpoint= true;
    }


    public void updateEyePosition()
    {
	distance_between_viewingplane_zCentre= lengthfactor * fixed_distance_between_viewingplane_zCentre;
	// update distance_betweenEye_LookAt
	distance_betweenEye_LookAt= readingplane_dvalue + distance_between_viewingplane_zCentre; //adding some constant
	// update eye position
	double z, y;
	z= distance_betweenEye_LookAt * Math.cos(alpha);
	y= distance_betweenEye_LookAt * Math.sin(alpha);
	eye.set(0.0, y, z);
    }


    /**
     * 
     */
    protected void initViewingParam()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("Configuration.java: initViewingParam()");
	}

	FOV= 45.0;
	updateAspect(windowbounds.width, windowbounds.height);
	//near= paperHeight/Math.tan(Utilities.degree2rad(FOV));
	near= zMax;
	
	
	//far= zMax * (Math.max(EYE_Z_FACTOR, EYE_Y_FACTOR) + 2);
	far= zMax * 20;
	centerZ= zMax / 2.0;
	//height= xMax / Math.tan( FOV*(Math.PI/180.0) );

	System.out.println("Configuration.java: aspect> " + aspect);
	System.out.println("Configuration.java: FOV> " + FOV);
	System.out.println("Configuration.java: far> " + far);
	System.out.println("Configuration.java: near> " + near);
	

	//predetermined eye
	eye= new Point3d(0.0, EYE_Y_FACTOR*(zMax+10), EYE_Z_FACTOR*(zMax+10));
	//eye= new Point3d(0.0, EYE_Y_FACTOR*(zMax+10), 10);
	lookAt= new Point3d(0.0, 0.0, centerZ);
	upVector= new Vector3d(0.0, 1.0, 0.0);

	aVector= new Vector3d(eye).normalized();
	bVector= new Vector3d(lookAt).normalized();
	double angle= Utilities.rad2degree(aVector.angle(bVector));
	howmuchtoturn= 90 - angle;

	System.out.println("Configuration.java: angle between aVector and bVector> " + angle);
	System.out.println("Configuration.java: howmuchtoturn> " + howmuchtoturn);
    }


    protected void initZoominParam()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("Configuration.java: initZoominParam()");
	}

	distance_betweenEye_LookAt= eye.distance(lookAt);
	System.out.println("Configuration.java: distance_betweenEye_LookAt> " + distance_betweenEye_LookAt);
	

    }

    
    public void updateReadingPlane_dvalue()
    {
	
	if(Ttp.commentslevel >= 2){
	    System.out.println("Configuration.java: updateReadingPlane_dvalue()");
	}

	

	if(readingplane_aspect > aspect){
	    if(Ttp.commentslevel >= 3){
		System.out.println("Configuration.java: based on width");
	    }
	    //based on win_width 
	    //recalculate height
	    double new_paperHeight= (paperWidth+paperWidth)/aspect;
	    double half_paperHeight= (new_paperHeight)/2;
	    double half_fov= FOV/2;
	    readingplane_dvalue= half_paperHeight/Math.tan(Utilities.degree2rad(half_fov));

	    

	}else{ // readingplane_aspect <= aspect
	    //based on height
	    if(Ttp.commentslevel >= 3){
		System.out.println("Configuration.java: based on height");
	    }
	    double half_paperHeight= (paperHeight)/2;
	    double half_fov= FOV/2;
	    readingplane_dvalue= half_paperHeight/Math.tan(Utilities.degree2rad(half_fov));
	    
	}
	    
	if(Ttp.commentslevel >= 3){
	    System.out.println("Configuration.java: readingplane_dvalue>  " + readingplane_dvalue);
	}
    }
    
    
    /**
     * 
     */
    public void updateAspect(int width, int height)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("Configuration.java: updateAspect()");
	}
		
	aspect= (double)width/(double)height;
	windowbounds.width= width;
	windowbounds.height= height;
	
	
    }

    
    
    /**
     * query command
     */
    public void getValue(int pname, double pvalue [])
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("Configuration.java: getValue()");
	}

	//System.out.println("Configuration.java: getValue(): ");
	switch(pname){
	case PAGE_DIMENSION:{
	    if(Ttp.commentslevel >= 3){
		System.out.println("page_dimention");
	    }
	    
	    pvalue[0]= (double)nx;
	    pvalue[1]= (double)nz;
	    pvalue[2]= elementLength;
	    break;
	}
	
	case COVER_DIMENSION:{ 
	    if(Ttp.commentslevel >= 3){
		System.out.println("cover_dimension");
	    }
	    pvalue[0]= zMax;
	    pvalue[1]= xMax;
	    break;
	}

	case PERSPECTIVE_PARAMETERS:{
	    if(Ttp.commentslevel >= 3){
		System.out.println("perspective_parameters");
	    }
	    pvalue[0]= FOV;
	    pvalue[1]= aspect;
	    pvalue[2]= near;
	    pvalue[3]= far;
	    break;
	}
	
	case CAMERA_PARAMETERS:{
	    if(Ttp.commentslevel >= 3){
		System.out.println("camera_parameters");
	    }

	    pvalue[0]= eye.x;       pvalue[1]= eye.y;       pvalue[2]= eye.z;
	    pvalue[3]= lookAt.x;    pvalue[4]= lookAt.y;    pvalue[5]= lookAt.z;
	    pvalue[6]= upVector.x;  pvalue[7]= upVector.y;  pvalue[8]= upVector.z;
	    
	    break;
	}

	case BOOK_PARAMETERS:{ 
	    if(Ttp.commentslevel >= 3){
		System.out.println("book_parameters");
	    }
	    
	    pvalue[0]= (double)pdf_total_pages;
	    
	    break;
	}

	case INTERSECTION_POINTS:{ 
	    if(Ttp.commentslevel >= 3){
		System.out.println("intersection_points");
	    }
	    
	    int p= 0;
	    for(int i=0 ; i<4; i++){
		p= i*2;
		pvalue[p]= points[i].x;
		pvalue[p+1]= points[i].y;
	    }
	    break;
	}

	
	default: // warning !!
	    if(Ttp.commentslevel >= 3){
		System.out.println("****Configuration.java: getValue() Unknown pvalue!!");
	    }
	}
	
    }

    public double getCoverWidth()
    {
	return cover_width;
    }

    public double getBookThickness()
    {
	return book_thickness;
    }

    public double getPaperThickness()
    {
	return paperThickness;
    }

    public double getCoverMargin()
    {
	return cover_margin;
    }

    public double getCoverThickness()
    {
	return cover_thickness;
    }
    
    public double getCoverRatio()
    {
	return cover_ratio;
    }
    
    public double get_zMax()
    {
	return zMax;
    }

    //testing
    public double get_rotCk2()
    {
	return rotAround_ck2;
    }
    

    //testing
    public int get_current_paper()
    {
	return current_paper;
    }

    public int get_total_papers()
    {
	return total_papers;
    }

    public int get_total_content_pages()
    {
	return total_content_pages;
    }

    public int get_thefirstPdfpage()
    {
	return thefirstPdfpage;
    }
    
    public int get_thefirstPngpage()
    {
	return thefirstPNGpage;
    }
    
    
    public int get_pdf_total_pages()
    {
	return pdf_total_pages;
    }
    
    public double getPaperHeight()
    {
	return paperHeight;
    }
    
    public double getPaperWidth()
    {
	return paperWidth;
    }
    
    //new 7_19
    public boolean isOK2LoadPositions()
    {
	return PR.isOK();
    }

     

    public int get_nx(){ return nx; }
    public int get_nz(){ return nz; }
    public double get_elementLength(){ return elementLength; }
    public double get_angleStep()
    {
	return PR.getAngleStep();
    }
    
    public double getMaxAngle()
    {
	return  PR.getMaxAngle();
    }
    
    public double getSpineface2_height()
    {
	return spinef2_height;
    }
    
    public double getSpineface5_height()
    {
	return spinef5_height;
    }

    public void setRegion(int region)
    {
	this.region= region;
	PR.initialize_interpolation(region);
    }

    public int getRegion()
    {
	return this.region;
    }

    public double getTimeStep()
    {
	return timeStep;
    }
    
    
    public double getdisplayInterval()
    {
	return displayInterval;
    }
   
    public double getTurningInterval()
    {
	return turningInterval;
    }
    
    public double getPausingInterval()
    {
	return pausing;
    }
    
    public double getangleStep()
    {
	return angleStep;
    }

    public float [] getlight0_position(){ return LIGHT0_POSITION;}
    public float [] getambient_light_colour(){ return AMBIENT_LIGHT_COLOUR;}
    public float [] getdiffuse_light_colour(){ return DIFFUSE_LIGHT_COLOUR;}
    public float [] getspecular_light_colour(){ return SPECULAR_LIGHT_COLOUR;}
    public float [] getlightmodel_ambient(){ return lmodel_ambient;}
    
    public float [] getPageColor(){return PAGE_COLOUR;}
    public float [] getPageSpecularColor(){return PAGE_SPECULAR;}
    public float getPageShiness(){return PAGE_SHINESS;}

    public float [] getCoverColor(){return COVER_COLOUR;}
    public float [] getCoverSpecularColor(){return COVER_SPECULAR;}
    public float getCoverShiness(){return COVER_SHINESS;}

    /* Set function*/
    public void setPageColor(Color4f newColor)
    {
	PAGE_COLOUR[0]= newColor.x;
	PAGE_COLOUR[1]= newColor.y;
	PAGE_COLOUR[2]= newColor.z;
	PAGE_COLOUR[3]= 0.8f;

    }

    
    private float calculateDeltaZ(double z0, double zLast, int levels)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("Configuration.java: calculateDeltaZ()");
	}

	float result;
	double x= zLast/z0;
	double y= 1/(double)(levels-1);
	result= (float)Math.pow(x, y) - 1.f;
	
	if(Ttp.commentslevel >= 3){
	    System.out.println("Configuration.java: calculateDeltaZ() x= " + x);
	    System.out.println("Configuration.java: calculateDeltaZ() y= " + y);
	    System.out.println("Configuration.java: calculateDeltaZ() deltaZ= " + result);
	}

	return result;
    }


    public int getInit_coverState()
    {
	return init_coverState;
    }

    public double getLowerZoomfactor(){ return this.lower_zoom_factor; }
    
    public Point3d getNewEyePosition()
    {
	return neweye;
    }

    

    
    
    
    
    
    

}
