// A FloatValueField is a subclass of JTextField
// that allows only valid floating point values
// to be entered at any time.

package TTPproject;

import javax.swing.*; 
import javax.swing.text.*; 
import java.awt.*;
import java.awt.event.ActionListener;

public class FloatValueField extends JTextField {

    ActionListener listener = null;
    int columns;
    
    public FloatValueField(float value, int columns, ActionListener listener) {
      super(columns);
      this.columns = columns;
      setDocument(new FloatNumberDocument());
      setValue(value);
      this.listener = listener;
      if (listener != null)
        addActionListener(listener);
    }
    
    public FloatValueField(float value, int columns) {
      this(value, columns, null);
    }

    public FloatValueField(float value, ActionListener listener) {
      this(value, 6, listener);
    }
    
    public FloatValueField(float value) {
      this(value, 6);
    }
    
    public float getValue() {
      float retVal = 0.0f;

      try {
          retVal = new Float(getText()).floatValue();
      } catch (NumberFormatException e) {
          // This should only happen if input field is an
          // isolated "-" sign, because insertString allows
          // only properly formatted data to get in the field.
      }
      return retVal;
    }

    public void setValue(float value) {
      setText("" + value);
    }
    
    public Dimension getPreferredSize() {
      return new Dimension(8 * columns, super.getPreferredSize().height);
    }
    
    public Dimension getMaximumSize() { return getPreferredSize(); }
}


class FloatNumberDocument extends PlainDocument  {

  public void insertString(int offs, String str, AttributeSet a)
                  throws BadLocationException {

    String currentText = getText(0, getLength());
    String beforeOffset = currentText.substring(0, offs);
    String afterOffset = currentText.substring(offs,
                                         currentText.length());
    String proposedResult = beforeOffset + str + afterOffset;

    try {
      if (!proposedResult.trim().equals("-"))  // Allow isolated "-"
        new Float(proposedResult).floatValue(); // Try converting to float
        super.insertString(offs, str, a);
    } catch (NumberFormatException e) {
        Toolkit.getDefaultToolkit().beep();
    }
  }    
}
