package TTPproject;

import java.io.*;// Random Access File
import java.util.*;
import javax.vecmath.*;// Point3d

public class ParticlesReader3 
    implements CONSTANTS{
    
    //constants
    public static final int INTSIZE= 4; // bytes

    File file= null;
    RandomAccessFile in= null;
    boolean ok= false;
    
    //header
    String code= "LobbVector";
    int nx=0 , nz=0 ;
    double meshelementlength= 0.0;
    double angleStep= 0.0;
    
    int max_nParticlesSets= 0;
    double biggest_angle= 0;

    class LobbsRecord{
	//fixed value
	int nx=0 , nz=0 ;
	double meshelementlength= 0.0;
	double angleStep= 0.0;

	//changing variables
	int nParticlesSets= 0;
	double max_angle= 0;
	List list= null;
    }

    //file pointer
    long fptr= 0;

    List list= null;
    String [] fname={"lobbvectorTop.ttp",
		     "lobbvectorMiddle.ttp",
		     "lobbvectorBottom.ttp"};
    LobbsRecord [] lobbsRecord= null;
    
    int region= TOP;
    
    int r1= region, r2= region;
    int half_region= 0; //used for seperate
    int total_no_regions= 0;
    int upper_no_regions, lower_no_regions;

    double beta;
    
    //Contructor
    //@param fname
    //open the file
    public ParticlesReader3(String prefix)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("ParticlesReader3.java: Reading Page turning data");
	}

	int i= 0;
	lobbsRecord= new LobbsRecord[fname.length]; 
	
	try{
	    while(i<fname.length){
		lobbsRecord[i]= new LobbsRecord();
		file= new File( prefix+fname[i] );// not very efficient 
		in= new RandomAccessFile(file, "r");
		if(readHeader(i)){
		    if(read_nParticlesSets(i)){
			readContents(i);
		    }
		}
		
		in.close();
		
		i++;
	    }

	    total_no_regions= (nz-1);
	    half_region= total_no_regions/2;
	    upper_no_regions= half_region-1;
	    lower_no_regions= (total_no_regions - half_region);
	    if(Ttp.commentslevel >= 3){
		System.out.println("ParticlesReader3: regions");
		System.out.println("ParticlesReader3: half_region " + half_region);
		System.out.println("ParticlesReader3: upper_no_regions " + upper_no_regions);
		System.out.println("ParticlesReader3: lower_no_regions " + lower_no_regions);
	    }
	    
	}catch(IOException e){
	    if(Ttp.commentslevel >= 3){
		System.out.println("ParticlesReader3: IOException: " + e);
	    }
	    ok= false;
	}

    }

    public boolean isOK()
    {
	return ok;
    }

    public boolean readHeader(int index)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("ParticlesReader3: readHeader");
	}

	StringBuffer buf= new StringBuffer(20);
	char chr; 
	//char lineSep= System.getProperty("line.separator").charAt(0);
	//System.out.println("System: line separator: " + lineSep);
	try{
	    //extract Code first
	    while ((chr = in.readChar()) != '\r')
		buf.append(chr);
	    //match
	    if(code.equalsIgnoreCase(buf.toString())){
		if(Ttp.commentslevel >= 3){
		    System.out.println(code + " is correct");
		}
		ok= true;
	    }else{
		if(Ttp.commentslevel >= 3){
		    System.out.println(code + " is not correct");
		}
		ok= false;
		return ok;
	    }

	    //read nx, nz, meshelementlength
	    nx= in.readInt();
	    nz= in.readInt();
	    
	    meshelementlength= in.readDouble();
	    angleStep= in.readDouble();
	    fptr= in.getFilePointer(); // save fptr

	    //debug
	    if(Ttp.commentslevel >= 3){
		System.out.println("\tnx> " + nx);
		System.out.println("\tnz> " + nz);
		System.out.println("\tmeshelementlength> " + meshelementlength);
		System.out.println("\tangle Step> " + angleStep);
		System.out.println("cur file ptr locates at > " + fptr);
	    }

	    
	    
	}catch(IOException e){
	    if(Ttp.commentslevel >= 3){
		System.out.println("****ParticlesReader3.java: readHeader() IOException: " + e);
	    }
	    ok= false;
	}
	
	return ok;
    }


    public void readContents(int index)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("ParticlesReader3: readContents");
	}
	
	Vector3d [] lobbvectors;
	int i, size= nx * nz ;
	
	
	for(i=0; i < lobbsRecord[index].nParticlesSets; i++){
	    //positions= new Point3d[size];
	    lobbvectors= new Vector3d[size];
	    //readPositions(positions);
	    readVectors(lobbvectors);
	    //list.add(i,(Point3d [])positions);
	    lobbsRecord[index].list.add(i, (Vector3d []) lobbvectors);
	    //debug
	    //displayPositions((Point3d [])list.get(i));
	    //displayVectors((Vector3d [])list.get(i));
	    
	}
    }
    
    void readVectors(Vector3d [] lobbvectors)
    {
	int i;
	if(Ttp.commentslevel >= 2){
	    System.out.println("ParticlesReader3: readVectors");
	}
	
	try{
	    for(i=0; i < lobbvectors.length; i++){
		lobbvectors[i]= new Vector3d();
		lobbvectors[i].x= in.readDouble();
		lobbvectors[i].y= in.readDouble();
		lobbvectors[i].z= in.readDouble();
	    }

	}catch( IOException e){
	    if(Ttp.commentslevel >= 3){
		System.out.println("****IOException: readPositions: " + e);
	    }
	    ok= false;
	}
    }
	

   

    public boolean read_nParticlesSets(int index)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("ParticlesReader3: read_nParticlesSets");
	}
	
	try{
	    //moving file pointer
	    in.seek(in.length() -  (long)INTSIZE);
	    //read nParticelsSets
	    lobbsRecord[index].nParticlesSets= in.readInt();
	    //debug
	    if(Ttp.commentslevel >= 3){
		System.out.println("ParticlesReader3: nParticlesSets>" + lobbsRecord[index].nParticlesSets);
	    }
	    
	    lobbsRecord[index].max_angle= 
		((lobbsRecord[index].nParticlesSets) - 1)* angleStep;
	    if(Ttp.commentslevel >= 3){
		System.out.println("ParticlesReader3: max_angle> " + lobbsRecord[index].max_angle);
	    }
	    
	    //  if(lobbsRecord[index].max_angle > biggest_angle){
	    //  		biggest_angle= lobbsRecord[index].max_angle;
	    //  		max_nParticlesSets= lobbsRecord[index].nParticlesSets;
	    
	    //  	    }

	    ok=true;
	    lobbsRecord[index].list= new ArrayList( lobbsRecord[index].nParticlesSets);
	    //moving file pointer back
	    in.seek(fptr);
	    if(Ttp.commentslevel >= 3){
		System.out.println("****ParticlesReader3: cur fptr locates at> " + fptr);
	    }
	    
	}catch(EOFException e){
	    if(Ttp.commentslevel >= 3){
		System.out.println("****EOFException: read_nParticlesSets");
	    }
	    ok= false;
	}
	catch(IOException e){
	    if(Ttp.commentslevel >= 3){
		System.out.println("****IOException: read_nParticlesSets");
	    }
	    ok= false;
	}
	
	return ok;
    }

    public void loadLobbsvector(int region, int row, double angle, Vector3d [] vectors)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("ParticlesReader3: loadLobbsvector");
	}
	
	Vector3d [] copyfrom= null;
	Vector3d [] copyfrom2= null;
	
	double index= 0;
	if(angle < 0) angle= 0;
	if(angle > biggest_angle) 
	    angle= biggest_angle;

	index= angle/ (biggest_angle) * (max_nParticlesSets-1);
	
	if(Ttp.commentslevel >= 3){
	    System.out.println("max_nParticleSets > " + max_nParticlesSets );
	    System.out.println("row > " + row );
	    System.out.println("angle > " + angle);
	    System.out.println("i > " + index );
	}
	
	//There is no way that r1 & r2 both are greater than 
	//nParticlesSets

	if(index >= lobbsRecord[r1].nParticlesSets && 
	   index >= lobbsRecord[r2].nParticlesSets){
	    if(Ttp.commentslevel >= 3){
		System.out.println("Something wrong");
	    }
	}

	if(index >= lobbsRecord[r1].nParticlesSets){
	    if(Ttp.commentslevel >= 3){
		System.out.println("r1 reach the end early");
	    }
	    copyfrom= (Vector3d []) (lobbsRecord[r1].list).get(lobbsRecord[r1].nParticlesSets-1); //return the whole set of vectors1
	    copyfrom2= (Vector3d []) (lobbsRecord[r2].list).get((int)index); //return the whole set of vectors2
	    
	}else if(index >= lobbsRecord[r2].nParticlesSets){
	    if(Ttp.commentslevel >= 3){
		System.out.println("r2 reach the end early");
	    }
	    copyfrom= (Vector3d []) (lobbsRecord[r1].list).get((int)index); //return the whole set of vectors2
	    copyfrom2= (Vector3d []) (lobbsRecord[r2].list).get(lobbsRecord[r2].nParticlesSets-1); //return the whole set of vectors1
	    
	}else{
	    copyfrom= (Vector3d []) (lobbsRecord[r1].list).get((int)index); //return the whole set of vectors1
	    copyfrom2= (Vector3d []) (lobbsRecord[r2].list).get((int)index); //return the whole set of vectors2
	}
	

	
	int i, n;
	for(i=0 ; i < nx; i++){ // extract the specific row of vectors from the specified set
	    n= i*nz + row;
	    vectors[i].interpolate(copyfrom[n], copyfrom2[n], beta);
	  
	}
    }

    //new
    public int getNx(){return nx;}
    public int getNz(){return nz;}
    public double getMeshElementLength(){return meshelementlength;}
    public double getAngleStep(){return angleStep;}
    public double getMaxAngle(){return biggest_angle;}
    
   
    public void initialize_interpolation(int region)
    {
	if(Ttp.commentslevel >= 3){
	    System.out.println("ParticleReader3.java: initialize_interpolation():");
	}

	this.region= region; // between 0 .. nz-1

	if(region < half_region){
	    if(Ttp.commentslevel >= 3){
		System.out.println("Top half is seleceted");
	    }
	    beta= (double) region/upper_no_regions;
	    
	    r1= TOP;
	    r2= MIDDLE;
	    
	    if(Ttp.commentslevel >= 3){
		System.out.println("region> " + region);
		System.out.println("upper_no_regions> " + upper_no_regions);
	    }
	}else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("Lower half is seleceted");
	    }
	    beta= (1.0 - (double) (region- half_region)/lower_no_regions);
	    
	    if(Ttp.commentslevel >= 3){
		System.out.println("region minus half> " + (region-half_region) );
		System.out.println("lower_no_regions> " + lower_no_regions);
	    }

	    r1= BOTTOM;
	    r2= MIDDLE;
	}
	
	//System.out.println("beta> " + beta);
	
	//  switch(region){
//  	case TOP:{ 
//  	    r1= region; 
//  	    r2= MIDDLE; 
//  	    break;
//  	}
//  	case BOTTOM:{ 
//  	    r1= region;
//  	    r2= MIDDLE; 
//  	    break;
//  	}
//  	case MIDDLE:{  
//  	    r1= region;
//  	    r2= TOP; 
//  	    break;
//  	}
//  	}

	if(lobbsRecord[r1].max_angle > lobbsRecord[r2].max_angle){
	    if(Ttp.commentslevel >= 3){
		System.out.println("r1 is greater than r2");
	    }
	    biggest_angle= lobbsRecord[r1].max_angle;
	    max_nParticlesSets= lobbsRecord[r1].nParticlesSets;
	}else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("r2 is greater than r1");
	    }

	    biggest_angle= lobbsRecord[r2].max_angle;
	    max_nParticlesSets= lobbsRecord[r2].nParticlesSets;
	}

    }

}
