/**
 * A utility class to load texture images from JPEG
 * files and converts them to OpenGL format (Magician style).
 * This is a simplified reimplementation of the Magician
 * texture class, which I could never get working!
 * 
 * @author Richard Lobb 
 * @version 26 July 2001.
 */
package TTPproject;

/*gl4java*/
import gl4java.utils.textures.*;
import gl4java.GLFunc;
import gl4java.GLUFunc;

import java.awt.Image;
import java.awt.image.*;
import com.sun.image.codec.jpeg.*;
import java.io.*;

public class PublicJPEGTextureLoader
extends IOTextureLoader
{

    protected boolean grayToAlpha = false;
    
    byte[][][] rgb;
    byte[] rgb1D;
    int tWidth, tHeight;

    public PublicJPEGTextureLoader(GLFunc gl, GLUFunc glu)
    {
	super(gl, glu);
    }


    /**
     * Set to 'true' to load grayscale images as alpha images
     * (in other words, GL_ALPHA instead of GL_LUMINANCE)
     */        
    public void setGrayToAlpha(boolean b)
    {
	this.grayToAlpha = b;
    }
    
    
    
    protected boolean readTexture(InputStream is){
	try {
            JPEGImageDecoder decoder = JPEGCodec.createJPEGDecoder(is);
            BufferedImage image = decoder.decodeAsBufferedImage();
	    
	    imageWidth = image.getWidth();
            imageHeight = image.getHeight();

            int width = image.getWidth();
            int height = image.getHeight();
            tWidth = bestPowerOfTwo(width);
            tHeight = bestPowerOfTwo(height);

	    if(Ttp.commentslevel >= 3){
		System.out.println("Original image size = " + width + " x " + height);
		System.out.println("Scaled image size = " + tWidth + " x " + tHeight);
	    }
	    
            Image scaledImage = (width == tWidth && height == tHeight) ? image :
                    image.getScaledInstance(tWidth, tHeight, Image.SCALE_SMOOTH);
            int[] rgbInts = new int[tWidth * tHeight];
            PixelGrabber grabber = new PixelGrabber(scaledImage, 0, 0, tWidth, tHeight,
                rgbInts, 0, tWidth);
            if(!grabber.grabPixels())
                throw new RuntimeException("GrabPixels returned false");
            rgb = new byte[tWidth][tHeight][4];
            rgb1D = new byte[tWidth * tHeight * 4];
            int index = 0;
            for (int y = 0; y < tHeight; y++) {
                for (int x = 0; x < tWidth; x++) {
                    int pixel = rgbInts[y * tWidth + x];
                    rgb1D[index++] = rgb[x][y][0] = (byte) ((pixel >> 16) & 0XFF);
                    rgb1D[index++] = rgb[x][y][1] = (byte) ((pixel >> 8) & 0XFF);
                    rgb1D[index++] = rgb[x][y][2] = (byte) (pixel & 0XFF);
                    rgb1D[index++] = rgb[x][y][3] = (byte) ((pixel >> 24) & 0XFF);
                 }
            }
         }
	catch (Exception e) { 
	    e.printStackTrace();
	    error= true;
	    return false;
	}
	
	return true;
	
    }

    public boolean publicreadTexture(InputStream is){
	return readTexture(is);
    }
    

    /**
     * Constructor for objects of class JPEGTextureLoader
     * @param filename file name (path) of a jpeg image file
     * to be used as the source of the texture. Since OpenGL
     * textures must be integer powers of two in size, the
     * texture that we build is resampled on a grid of size
     * 2^n x 2^m where n and m are integers that give the
     * nearest fit to the image width and height resp.
     */

    /*public JPEGTextureLoader(String filename)
    {
        try {
            JPEGImageDecoder decoder = JPEGCodec.createJPEGDecoder(new FileInputStream(filename));
            BufferedImage image = decoder.decodeAsBufferedImage();
            int width = image.getWidth();
            int height = image.getHeight();
            tWidth = bestPowerOfTwo(width);
            tHeight = bestPowerOfTwo(height);
            // System.out.println("Original image size = " + width + " x " + height);
            // System.out.println("Scaled image size = " + tWidth + " x " + tHeight);
            Image scaledImage = (width == tWidth && height == tHeight) ? image :
                    image.getScaledInstance(tWidth, tHeight, Image.SCALE_SMOOTH);
            int[] rgbInts = new int[tWidth * tHeight];
            PixelGrabber grabber = new PixelGrabber(scaledImage, 0, 0, tWidth, tHeight,
                rgbInts, 0, tWidth);
            if(!grabber.grabPixels())
                throw new RuntimeException("GrabPixels returned false");
            rgb = new byte[tWidth][tHeight][4];
            rgb1D = new byte[tWidth * tHeight * 4];
            int index = 0;
            for (int y = 0; y < tHeight; y++) {
                for (int x = 0; x < tWidth; x++) {
                    int pixel = rgbInts[y * tWidth + x];
                    rgb1D[index++] = rgb[x][y][0] = (byte) ((pixel >> 16) & 0XFF);
                    rgb1D[index++] = rgb[x][y][1] = (byte) ((pixel >> 8) & 0XFF);
                    rgb1D[index++] = rgb[x][y][2] = (byte) (pixel & 0XFF);
                    rgb1D[index++] = rgb[x][y][3] = (byte) ((pixel >> 24) & 0XFF);
                 }
            }
         }
         catch (Exception e) { e.printStackTrace(); }
	 }*/

    /**
     * Get width of actual texture. 
     * 
     * @return     width of texture (in pixels). This will always be
     * a power of two and is the nearest power of two to the
     * width of the image texture loaded originally.
     */
    public int getWidth() { return tWidth; }


    /**
     * Get height of actual texture. 
     * 
     * @return     height of texture (in pixels). This will always be
     * a power of two and is the nearest power of two to the
     * height of the image texture loaded originally.
     */
    public int getHeight() { return tHeight; }

    /** 
     * Get texture array in Magician/OpenGL compatible format.
     * @return  the texture as a byte array. First subscript
     * is the column (x), second is the row (y), third is 0 .. 3 for 
     * red, green, blue and alpha respectively.
     */

    public byte[][][] getTextureRGB() { return rgb; }

    public byte[] getTexture1D() { return rgb1D; }
    

    private int bestPowerOfTwo(int size) {
        int ans = 1;
        while (ans * 2 < size) ans *= 2;
        if (2 * ans - size < size - ans) ans *= 2;
        return ans;
    }

    private byte[] getPixel(BufferedImage image, float x, float y) {
        // Returns {R,G,B,255} colour of given image at relative position (x,y).
        // Uses bilinear interpolation.
        float fx0 = (x * (image.getWidth() - 1));
        float fy0 = (y * (image.getHeight() - 1));
        int x0 = (int) fx0;
        int y0 = (int) fy0;
        byte[] c00 = getAndUnpackRGB(image, x0, y0);
        byte[] c10 = getAndUnpackRGB(image, x0 + 1, y0);
        byte[] c01 = getAndUnpackRGB(image, x0, y0 + 1);
        byte[] c11 = getAndUnpackRGB(image, x0 + 1, y0 + 1);
        float dx = fx0 - x0;
        float dy = fy0 - y0;
        byte[] result = new byte[4];
        for (int i = 0; i < 3; i++)
            result[i] = bilinearInterp(c00[i], c10[i], c01[i], c11[i], dx, dy);
        result[3] = (byte) 255;
        return result;
    }

    private byte[] getAndUnpackRGB(BufferedImage image, int x, int y) {
        x = Math.min(x, image.getWidth() - 1);
        y = Math.min(y, image.getHeight() - 1);
        int rgb = image.getRGB(x,y);
        return new byte[] {
            (byte) ((rgb >> 16) & 255),
            (byte) ((rgb >> 8) & 255),
            (byte) (rgb & 255)};
    }

    private byte bilinearInterp(int c00, int c10, int c01, int c11, float dx, float dy) {
        c00 &= 255; c10 &= 255; c01 &= 255; c11 &= 255;
        float w00 = (1 - dx) * (1 - dy);
        float w10 = dx * (1 - dy);
        float w01 = (1 - dx) * dy;
        float w11 = dx * dy;
        int result = (int) (w00 * c00 + w01 * c01 + w10 * c10 + w11 * c11);
        return (byte) Math.min(255,result);
    }
}
