package TTPproject;

import gl4java.awt.GLCanvas;
import gl4java.GLContext;
import gl4java.GLFunc;
import gl4java.GLUFunc;
import gl4java.GLEnum;

import java.util.Map;
import java.util.HashMap;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

import java.util.StringTokenizer;
import java.util.List;
import java.util.ArrayList;// anchorPageList

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;


/*
  Maintains a list of texture names and perform texture
  generation process

  It has to interface with Texture

  Data Structure 
  Queue FIFO behaviour

  HashMap, (Note that this implementation is not synchronized, merely changing the value 
  associated with a key that an instance already contains is not structural modification)

  Page starts from 0 to last page 

  Cover 0 1 2 3


 */
public class TexturePools implements CONSTANTS{
      
    class Page{
	private int texName; // a single texture  ??
	private byte [] imageBuffer;
	private int imageWidth;
	private int imageHeight;
	private int level;
	private int textureLevel; //page material

	/*
	  default constructor
	 */
	public Page(){
	    if(Ttp.commentslevel >= 2){
		System.out.println("TexturePools.java: Page() default");
	    }
	    
	    setTexName(0);//page texName 
	    setTextureLevel(0); //page material index, white page no one look at it
	    setImageBuffer(null);//page content
	    setImageWidth(0);
	    setImageHeight(0);
	    setLevel(0); //image resolution level indicator for progressive update
	}

	
	public Page(byte [] imageBuffer,
		    int level,
		    int w, int h){
	    
	    if(Ttp.commentslevel >= 2){
		System.out.println("TexturePools.java: Page() 3 arguments");
	    }
	    
	    setTexName(0);
	    setImageBuffer(imageBuffer);
	    setLevel(level);
	    setImageWidth(w);
	    setImageHeight(h);
	    setTextureLevel(1);
	}

	public void setTexName(int texName){this.texName=texName; }
	public void setImageBuffer(byte [] imageBuffer){this.imageBuffer= imageBuffer; }
	public void setLevel(int level){ this.level=level; }
	public void setImageWidth(int w){ this.imageWidth= w; }
	public void setImageHeight(int h){ this.imageHeight= h; }
	public void setTextureLevel(int tLevel){this.textureLevel= tLevel; }
	
	public void incrLevel(int v){ (this.level) += v;}
	
	public int getTexName(){return texName; }
	public byte [] getImageBuffer(){return imageBuffer; }
	public int getLevel(){return level; }
	public int getImageWidth(){return imageWidth; }
	public int getImageHeight(){return imageHeight; }
	public int getTextureLevel(){return textureLevel; }
    }

     

    private Map cache; 
        
    /*private GLCanvas canvas;
    private GLContext glj;
    private GLFunc gl;
    private GLUFunc glu;
    private Texture textureReader;*/
    private Configuration config= null;
    private Parameters parameters= null;
    private int from= 0;
    private int to= 1;
    private List anchorListRight= null; 
    private List anchorListLeft= null; 
    

    public TexturePools(int cache_size, 
			Parameters parameters,
			List anchorListLeft,
			List anchorListRight)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: TexturePools()");
	}
	
	this.cache= new HashMap(cache_size, .75f);
	this.config= config;
	this.parameters= parameters;
	this.from= parameters.getpageFrom();
	this.to= parameters.getpageTo();
	this.anchorListRight= anchorListRight;
	this.anchorListLeft= anchorListLeft;
	
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java:" );
	    System.out.println("TexturePools.java: from= " + from );
	    System.out.println("TexturePools.java: to= " + to );
	}
	
    }


    public void doInitialized(Configuration config)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: doInitialized()");
	}
	
	this.config= config;

	if(Ttp.commentslevel >= 3){
	    System.out.println("TexturePools.java: lastPage> " + config.lastPage);
	}
	
	//initialize page cache & create Page object
	int page_no= 0;
	while(page_no <= config.lastPage ){
	    initializePage(page_no);
	    page_no++;
	}

	readingPreCompilePages();
	setNoizelevel();
	readingAnchorPages();
    }


    protected void readingPreCompilePages()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: readingPreCompilePages()");
	}

	

	/*reading image pages from the data directory*/
	int page_no= from;
	while(page_no <= to ){
	    if(Ttp.commentslevel >= 3 ){
		System.out.println("TexturePools.java: page_no= " + page_no );
		System.out.println("TexturePools.java: maxlevel= " + config.maxlevel);
	    }
	    
	    putTexturefromFile(page_no, 
			       config.maxlevel, 
			       config.imageWidth_highresol, 
			       config.imageHeight_highresol);
	    page_no++;
	}

	Ttp.messagesState= 2;


    
    }

    protected void setNoizelevel()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: setNoizelevel()");
	}

	BufferedReader in;
	
	//read from log file
	try{
	    if(Ttp.commentslevel >= 3){
		System.out.println("TexturePools.java: Page noize level log file: " +  config.page_log_file);
	    }
	    
	    in = new BufferedReader(new FileReader(config.page_log_file));
	    String line;
	    	    
	    while( (line=in.readLine()) != null){
		if(Ttp.commentslevel >= 3){
		    System.out.println("TexturePools.java: current line> " + line );
		}
		
		//extract pagefrom pageTo texLevel
		int pageFrom, pageTo, texLevel;
		StringTokenizer st = new StringTokenizer(line);
		pageFrom= Integer.parseInt(st.nextToken());
		pageTo= Integer.parseInt(st.nextToken());
		texLevel= Integer.parseInt(st.nextToken());
		
		if(Ttp.commentslevel >= 3){
		    System.out.println("TexturePools.java: run() (pageFrom, pageTo, noize level)= (" +
				       pageFrom + ", "+ 
				       pageTo + ", "+ 
				       texLevel + ")" );
		}

		int page_no= pageFrom;
		while(page_no <= pageTo){
		    setPageMaterialLevel(page_no, texLevel);
		    page_no++;
		}
	    }
	    in.close();
	}catch(IOException e){
	    if(Ttp.commentslevel >= 3 ){
		System.out.println("****TexturePools.java: IOException: " + e);
	    }
	}


    }

    protected void readingAnchorPages()
    {
	

	int right, left;
	int texName;
	int level, imageWidth, imageHeight;
	int interval= 2; //has to be even number
	int firstPage= config.firstPage; 
	int lastPage= config.lastPage; 
	
	    	    
	right= 0;
	left= -1;
	
	
	String anchors= null;
	
	//reading anchor file
	//assuming the page number is in correct sequence
	try{
	    if(Ttp.commentslevel >= 2){
		System.out.println("TexturePools.java: reading Anchor Pages from .ash file");
	    }

	    
	    BufferedReader in
		= new BufferedReader(new FileReader(config.anchoring_file));
	    
	    anchors= in.readLine();// read one line
	    in.close();
	    
	    int curPage;
	    StringTokenizer st = new StringTokenizer(anchors);
	    while (st.hasMoreTokens()) {// threadD only stops when the program is terminated.
		curPage= Integer.parseInt(st.nextToken());
		if(Ttp.commentslevel >= 3){
		    System.out.println("TexturePools.java: run() Anchored page= " + curPage );
		}
		
		//check out of boundry
		if(curPage > lastPage || curPage < firstPage){
		    if(Ttp.commentslevel >= 3 ){
			System.out.println("TexturePools.java: run() Error curPage is out of boundry " );
		    }
		    
		    break;
		    }
		
		if( (curPage % 2)== 0 ){
		    right= curPage;
		    left= curPage-1;
		}else{
		    //if left is the lastPage
		    // right must greater than the lastPage
		    left= curPage;
		    right= curPage+1;
		}
			
		//right page
		if(Ttp.commentslevel >= 3 ){
		    System.out.println("TexturePools.java: run() anchor right page= "  + right);
		}
		
		//if PDF file
		if(parameters.getfileType() == PDF){
		    if(Ttp.commentslevel >= 3 ){
			System.out.println("TexturePools.java: run() PDF");
		    }
		    
		    //reading from file first
		    if(!putTexturefromFile(right,  // if not successful
					   config.maxlevel, 
					   config.imageWidth_highresol, 
					   config.imageHeight_highresol)){
			//reading from pdfreader
			putTexture(right,
				   1,
				   config.imageWidth_lowresol,
				   config.imageHeight_lowresol
				   );
		    }
		}else if(parameters.getfileType() == DJVU){
		    if(Ttp.commentslevel >= 3 ){
			System.out.println("TexturePools.java: run() DJVU");
		    }
		    
		    putTexturefromFile(right,  
				       config.maxlevel, 
				       config.imageWidth_highresol, 
				       config.imageHeight_highresol);
		    
		}else{
		    if(Ttp.commentslevel >= 3 ){
			System.out.println("****TexturePools.java: run() illegal type");
			}
		}
		//else if DVJU file 
		//reading from file
		
		//add to right anchor page list
		synchronized(anchorListRight) {
		    anchorListRight.add(new Integer(right));//Appends the specified element to the end of this list
		}
		
		//left page
		if(Ttp.commentslevel >= 3 ){
		    System.out.println("TexturePools.java: run() anchor left page= "  + left);
		}
		//if PDF  file
		if(parameters.getfileType() == PDF){
		    if(Ttp.commentslevel >= 3 ){
			System.out.println("TexturePools.java: run() PDF");
		    }
		    //reading from file first
		    if(!putTexturefromFile(left,  // if not successful
					   config.maxlevel, 
					   config.imageWidth_highresol, 
					   config.imageHeight_highresol)){
			//reading from pdfreader
			putTexture(left,
				   1,
				   config.imageHeight_lowresol,
				   config.imageWidth_lowresol
				   );
		    }
		}else if(parameters.getfileType() == DJVU){
		    if(Ttp.commentslevel >= 3 ){
			System.out.println("TexturePools.java: run() DJVU");
		    }
		    
		    putTexturefromFile(left,  
				       config.maxlevel, 
				       config.imageWidth_highresol, 
				       config.imageHeight_highresol);
		}else{
		    if(Ttp.commentslevel >= 3 ){
			System.out.println("****TexturePools.java: run() illegal type");
		    }
		}
		
		//add to left anchor page list
		synchronized(anchorListLeft) {
		    anchorListLeft.add(new Integer(left));//Appends the specified element to the end of this list
		}
	    }//end of while
	}catch(IOException e){
	    if(Ttp.commentslevel >= 3 ){
		System.err.println("****TexturePools.java: IOException: " + e);
	    }
		
	    right= 0;
	    left=-1;
	    int segments= 10;
	    //calculate interval based on total_content_pages
	    interval= config.total_content_pages/ segments;
	    
	    if(Ttp.commentslevel >= 3 ){
		System.out.println("****Caculate anchor page with interval: " + interval );
	    }
	    
	    if( (interval % 2) == 0 ){
		//even interval 
		if( interval == 0 )
		    interval = 2; //at least 2
	    }else{
		//odd interval 
		interval += 1;
	    }
	    
	    if(Ttp.commentslevel >= 3 ){
		System.out.println("TexturePools.java: run() interval="  + interval);
	    }
	    
	    
	    while(left <= lastPage){
		//change to even number 
		//right
		if(Ttp.commentslevel >= 3 ){
		    System.out.println("TexturePools.java: run() right anchor page= "  + right);
		}
		
		//reading from file first
		if(!putTexturefromFile(right,  // if not successful
				       config.maxlevel, 
				       config.imageWidth_highresol, 
				       config.imageHeight_highresol)){
		    //reading from pdfreader
		    putTexture(right,
			       1,
			       config.imageWidth_lowresol,
			       config.imageHeight_lowresol
			       );
		}
		
		
		//add to the right hand anchored page list
		synchronized(anchorListRight) {
		    anchorListRight.add(new Integer(right));//Appends the specified element to the end of this list
		}
		
		right += interval; 
		
		//left
		if(Ttp.commentslevel >= 3 ){
		    System.out.println("TexturePools.java: run() left anchor page= "  + left);
		    }
		
		if(left >= firstPage){
		    //reading from file first
		    if(!putTexturefromFile(left,  // if not successful
					   config.maxlevel, 
					   config.imageWidth_highresol, 
					   config.imageHeight_highresol)){
			//reading from pdfreader
			putTexture(left,
				   1,
				   config.imageHeight_lowresol,
				   config.imageWidth_lowresol
				   );
		    }

		    //add to the left hand anchored page list
		    synchronized(anchorListLeft) {
			anchorListLeft.add(new Integer(left));//Appends the specified element to the end of this list
		    }
		}
		left += interval;
	    }//end of while
	}//end of IOException
	

    }

  

    /*
      Generate image data from pdfreader
    */
    public synchronized boolean putTexture(int page_no, 
					   int level,
					   int imageWidth,
					   int imageHeight)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: putTexture(): " + page_no);
	}
	
	if(Ttp.commentslevel >= 3){
	    System.out.println("TexturePools.java: imageWidth> " + imageWidth);
	    System.out.println("TexturePools.java: imageHeight> " + imageHeight);
	}
	
	int pdf_name;
	byte [] imagebuffer;
	boolean success= false;

	Page curPage= (Page)(cache.get(new Integer(page_no)));
	
	//check curPage is not null
	if(curPage != null){
	    //receive image bytes from pdfreader
	    pdf_name= page_no + config.get_thefirstPdfpage();
	    imagebuffer= config.converter.getPageBytes(pdf_name,
						       imageWidth, 
						       imageHeight);
	    
	    //check imagebuffer contains data
	    if(imagebuffer != null){
		curPage.setImageWidth(imageWidth); 
		curPage.setImageHeight(imageHeight);
		curPage.setLevel(level);
		curPage.setImageBuffer(imagebuffer);
		success= true;
	    }else{
		if(Ttp.commentslevel >= 3){
		    System.out.println("****TexturePools.java: putTexture(): imagebuffer == null");
		}
		success= false;
	    }
	}else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("****TexturePools.java: putTexture(): Page object== null");
	    }
	    success= false;
	}

	return success;
    }

    


    //get page texture names
    public synchronized int [] getPageNames()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: getPageNames()");
	}
	
	Vector v= new Vector(cache.values());
	Object [] pagesArray = v.toArray();
	int size= pagesArray.length;
	int [] texNames= new int[size];

	for(int i= 0; i < size; i++){
	    texNames[i]= ((Page)pagesArray[i]).getTexName();
	}
	return texNames;
    }



    /*
      given page_no
      create image bytes
      insert into the cache
    */
    public synchronized boolean putTexturefromFile(int page_no,
						   int level,
						   int imageWidth,
						   int imageHeight)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: putTexturefromFile() page_no: " + page_no);
	}
	
	
	int png_page;
	byte [] imagebuffer;
	boolean success= false;
	Page curPage= (Page)(cache.get(new Integer(page_no)));
	
	if(curPage != null){
	    if(Ttp.commentslevel >= 3){
		System.out.println("TexturePools.java: putTexturefromFile(): Page object has initialized");
	    }
	    
	    png_page= page_no + config.get_thefirstPngpage();
	    
	    if(curPage.getLevel() == 0){ //has not initialized before
		//get image bytes from the file
		imagebuffer = config.converter.getPageBytesfromFile(png_page);
		
		if(imagebuffer != null){
		    if(Ttp.commentslevel >= 3){
			System.out.println("TexturePools.java: read in page image");
		    }
		    
		    curPage.setImageBuffer(imagebuffer);
		    curPage.setImageWidth(config.imageWidth_highresol);
		    curPage.setImageHeight(config.imageHeight_highresol);
		    curPage.setLevel(config.maxlevel);
		    success= true;
		}else{
		    if(Ttp.commentslevel >= 3){
			System.out.println("****TexturePools.java: putTexturefromFile(): file not existed");
		    }
		}
	    }
	    else{
		if(Ttp.commentslevel >= 3){
		    System.out.println("****TexturePools.java: putTexturefromFile(): page has been read");
		}
		success= true;
	    }
	    
	}else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("****TexturePools.java: putTexturefromFile(): Error Page not initialized");
	    }
	}
	
	return success;
    }

       

    /*
      initialize page
      Create Page object

     */
    public synchronized void initializePage(int page_no)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: initializePage()");
	}
	
	Page newPage= new Page();
	cache.put(new Integer(page_no), newPage);// page number is a key and texName is a value
    }

   

    /*
      Receive Page object from 'cache'
      calling converter.getPagesBytes() to generate image bytes from pdfreader
     */
    public synchronized void newPageTexture(int page_no)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: newPageTexture()");
	}
		    
	Page curPage= (Page)(cache.get(new Integer(page_no)));
	
	if(curPage != null){
	    if(curPage.getLevel() == 0){ //since default is 0
		byte [] imagebuffer;
		int pdf_name= page_no + config.get_thefirstPdfpage();
		
		//receive image bytes
		imagebuffer = config.converter.getPageBytes(pdf_name, 
							    config.imageWidth_lowresol, 
							    config.imageHeight_lowresol);

		//all precompiled images are at highest resolution
		if(imagebuffer != null){
		    if(Ttp.commentslevel >= 3){
			System.out.println("TexturePools.java: newPageTexture(): file not existed");
		    }
		    curPage.setImageBuffer(imagebuffer);
		    curPage.setImageWidth(config.imageWidth_lowresol);
		    curPage.setImageHeight(config.imageHeight_lowresol);
		    curPage.setLevel(1);
		}else{
		    //no such page
		    if(Ttp.commentslevel >= 3){
			System.out.println("TexturePools.java: newPageTexture(): imagebuffer is null");
		    }
		}
	    }
	}
	else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("****TexturePools.java: newPageTexture(): Error curPage == null ");
	    }
	}
	
    }

    public synchronized void setPageMaterialLevel(int page_no, int texLevel)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: setPageMaterialLevel()");
	}
	
	Page curPage= (Page)(cache.get(new Integer(page_no)));
	
	if(curPage != null){
	    curPage.setTextureLevel(texLevel);
	}
    }

    /*return level*/
    public synchronized int getPageMaterialLevel(int page_num)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: getPageMaterialLevel()");
	}
	
	Page curPage= (Page)(cache.get(new Integer(page_num))) ;
	if(curPage == null) 
	    return 0;
	else
	    return (curPage.getTextureLevel());
    }
    
    
    public synchronized void setTexName(int page_num, int newtexName)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: setTexName()");
	}
	
	int oldTexName;
	Page curPage= (Page)(cache.get(new Integer(page_num)));

	curPage.setTexName(newtexName);
    }
    
    public synchronized int getTexName(int page_num)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: getTexName()");
	}

	Page curPage= (Page)(cache.get(new Integer(page_num))) ;
	if(curPage == null) 
	    return 0;
	else
	    return (curPage.getTexName());
    }

    public synchronized boolean reachMaxLevel(int page_num)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: reachMaxLevel()");
	}

	Page curPage= (Page)(cache.get(new Integer(page_num))) ;
	if(curPage == null){
	    return false; //new texture
	}else{
	    if( curPage.getLevel() <= config.maxlevel )
		return true;
	    else
		return false;
	}
    }
    

    public synchronized boolean pageAlreadyExisted(int page_num)
    {
	return cache.containsKey(new Integer(page_num));
    }
    
    public synchronized boolean pageAlreadyinitialized(int page_no)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: pageAlreadyinitialized()");
	}
	
	boolean result=false;
	Page curPage= (Page)(cache.get(new Integer(page_no)));
	
	if(curPage != null){
	    if(curPage.getLevel() == 0){
		result= false;
	    }else{
		result= true;
	    }
	}else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("****TexturePools.java: pageAlreadyinitialized(): error curPage == null ");
	    }
	    result= false; // we think this as has been initialized
	}
	
	return result;
    }

    
    public synchronized void newPage(int page_num,
				     int level,
				     int imageWidth,
				     int imageHeight)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: newPage():");
	}
	
	int pdf_name;
	byte [] imagebuffer;

	//generate image bytes
	pdf_name= page_num + config.get_thefirstPdfpage();
	imagebuffer = config.converter.getPageBytes(pdf_name,
						    imageWidth,
						    imageHeight);
	
	Page newPage= new Page(imagebuffer, 
			       level,
			       imageWidth,
			       imageHeight);
	
	cache.put(new Integer(page_num), newPage);// page_no is a key and texName is a value
	
	if(Ttp.commentslevel >= 3){
	    System.out.println("TexturePools.java: putTexture(): imageWidth= " + newPage.getImageWidth());
	    System.out.println("TexturePools.java: putTexture(): imageHeight= " + newPage.getImageHeight());
	    System.out.println("TexturePools.java: putTexture(): level= " + newPage.getLevel());
	}
    }
    
    /*
      It is called during the initialization
      add the image with lowest resolution
    */
    /*public synchronized void putTexture2(int page_no)
    {
	System.out.println("TexturePools.java: putTexture2(): START");
	System.out.println("TexturePools.java: putTexture2(): page no= " + page_no);
	
	if(!cache.containsKey(new Integer(page_no))){
	    System.out.println("TexturePools.java: putTexture2(): not existed");
	    
	    //generate a new texName
	    int [] texName= new int[1];
	    gl.glGenTextures(1, texName);
	    System.out.println("TexturePools.java: putTexture2(): texture name= " + texName[0]);

	    Page newPage= new Page(texName[0], 
				   1,
				   config.imageWidth_lowresol,
				   config.imageHeight_lowresol);
	    
	    //add to the cache
	    System.out.println("TexturePools.java: putTexture2(): add to the cache ");
	    cache.put(new Integer(page_no), newPage);// page_no is a key and texName is a value
	    
	    //read texture from ...
	    int pdf_name= page_no + config.get_thefirstPdfpage();
	    if(textureReader.readPageTextureFromPdfReader(pdf_name, newPage.getImageWidth(), newPage.getImageHeight())){
		System.out.println("TexturePools.java: putTexture2(): Successful to read from the pdf, Continue to read the next one" );
		textureReader.setupMipmapsTexturing(texName[0]);
	    }else{
		System.out.println("TexturePools.java: putTexture2(): Fail to read from the pdf, Continue to read the next one" );
		//break;
	    }
	}else{
	    System.out.println("TexturePools.java: putTexture2(): already existed");
	    
	}

	System.out.println("TexturePools.java: putTexture2(): END");
	}*/

    public synchronized int getTexture(int page_no)  
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: getTexture()");
	    System.out.println("TexturePools.java: getTexture() " + page_no);
	}

	//Integer value= (Integer) (cache.get(new Integer(page_no))) ;
	Page curPage= (Page)(cache.get(new Integer(page_no))) ;
	int result= 0;
	if(curPage == null){
	    if(Ttp.commentslevel >= 3){
		System.out.println("****TexturePools.java: getTexture(): error did not find current page");
	    }
	}else{
	    //result= value.intValue();
	    result= curPage.getTexName();
	    if(Ttp.commentslevel >= 3){
		System.out.println("TexturePools.java: getTexture(): find the value= " + result);
	    }
	}
	
	
	return result;
    }

    public synchronized int getCurLevel(int page_no)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: getCurLevel() ");
	    System.out.println("TexturePools.java: getCurLevel() " + page_no);
	}
	
	int level;
	Page curPage= (Page)(cache.get(new Integer(page_no))) ;
	if(curPage == null){
	    if(Ttp.commentslevel >= 3){
		System.out.println("TexturePools.java: getCurLevel(): Do not find the page");
	    }
	    
	    level= 0;
	}else{
	    level= curPage.getLevel();
	}
	
	return level;
    }


    public synchronized int getCurHeight(int page_no)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: getCurHeight() START");
	    System.out.println("TexturePools.java: getCurHeight() " + page_no);
	}
	
	int height;
	Page curPage= (Page)(cache.get(new Integer(page_no))) ;
	if(curPage == null){
	    if(Ttp.commentslevel >= 3){
		System.out.println("TexturePools.java: getCurHeight(): Do not find the page");
	    }
	    
	    height= 0;
	}else{
	    height= curPage.getImageHeight();
	}
	
	return height;
    }


    public synchronized int getCurWidth(int page_no)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: getCurWidth() START");
	    System.out.println("TexturePools.java: getCurWidth() " + page_no);
	}
	
	int width;
	Page curPage= (Page)(cache.get(new Integer(page_no))) ;
	if(curPage == null){
	    if(Ttp.commentslevel >= 3){
		System.out.println("TexturePools.java: getCurWidth(): Do not find the page");
	    }

	    width= 0;
	}else{
	    width= curPage.getImageWidth();
	}

	return width;
    }
    
    
    public synchronized byte [] getImageBytes(int page_no)
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: getImageBytes() START");
	    System.out.println("TexturePools.java: getImageBytes() page_no= " + page_no);
	}
	
	//Integer value= (Integer) (cache.get(new Integer(page_no))) ;
	Page curPage= (Page)(cache.get(new Integer(page_no))) ;
	byte [] result= null;
	if(curPage == null){
	    if(Ttp.commentslevel >= 3){
		System.out.println("TexturePools.java: getImageBytes(): Do not find the page");
	    }
	}else{
	    result= curPage.getImageBuffer();
	}

	return result;
    }  

    

    public synchronized void displayCache()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("TexturePools.java: displayCache()");
	}
	//keySet() return a 'Set' of keys
	//values() returns a 'Collections' of values
	Vector v= new Vector(cache.values());
	if(Ttp.commentslevel >= 3){
	    System.out.println("Capacity>  " + v.capacity());
	}
	Set s= new HashSet(cache.keySet());
	
	Iterator the_iterator= s.iterator();
	if(Ttp.commentslevel >= 3){
	    System.out.print("Keys are[ ");
	    while(the_iterator.hasNext()){
		System.out.print(((Integer)(the_iterator.next())).intValue() + " ");
	    }
	    System.out.println("]");
	}

	/*Enumeration the_list = v.elements();
	int i=0; 
	while (the_list.hasMoreElements()){
	    System.out.println("Values["+ i +"]= " + ((Integer)(the_list.nextElement())).intValue());
	    i++;
	    }*/
		
	Object [] array = v.toArray();
	if(Ttp.commentslevel >= 3){
	    System.out.print("Values are[ ");
	    showArray(array);
	    System.out.println("]");
	}
    }


    void showArray(Object[] array )
    {
	for(int i = 0; i < array.length;
                                  i++){
	    if(array[i] == null){
		System.out.print("null ");
	    }else{
		System.out.print(((Page)array[i]).getTexName() + " ");
	    }//end else
	}//end for loop


    }


}
