package TTPproject;

import java.io.*;

import gl4java.awt.GLCanvas;
import gl4java.GLContext;
import gl4java.GLFunc;
import gl4java.GLUFunc;
import gl4java.GLEnum;

/*
  Animating the current right and left page 

 */
public class ThreadA extends Thread 
    implements GLEnum, CONSTANTS{
    private TexturePools texturePools;
    private int startPage;
    private Scene scene; 
    private int frontCoverPageTexName;
    
    /*gl4java*/
    private GLContext glj;
    private GLFunc gl;
    private GLUFunc glu;
        
    private boolean stop;
    private boolean threadSuspended;
    private Texture textureReader;
    private CursorManager cursorManager;
    
    /*static page*/
    public ThreadA(Scene scene,
		   CursorManager cursorManager,
		   TexturePools t,
		   int startPage)
    {
	super("threadA");
	if(Ttp.commentslevel >= 2){
	    System.out.println("ThreadA.java:" );
	}
	
	this.scene= scene;
	this.cursorManager= cursorManager;
	this.texturePools= t;
	this.stop= false;
	this.threadSuspended= false; //'false' means running, 'true' means waiting
	this.startPage= startPage;
    }
    

    /*
      put one texture into TexturePools
     */
    public void run()
    {
	
	if(Ttp.commentslevel >= 3){
	    System.out.println("ThreadA.java: run() START" );
	}
	

	this.glj= scene.getGLContext();
	this.gl= scene.getGL();
	this.glu= scene.getGLU();

	textureReader= new Texture(gl,
				   glu,
				   scene.config);

	
	int right, left;
	int texName;
	
	if( (startPage % 2)== 0 ){
	    right= startPage;
	    left= startPage-1;
	}else{
	    left= startPage;
	    right= startPage+1;
	}
	if(Ttp.commentslevel >= 3){
	    System.out.println("ThreadA.java: run() right="  + right);
	    System.out.println("ThreadA.java: run() left="  + left);
	}
	
	//starts with left hand page
	scene.setCursor(cursorManager.getCursor(WAIT));
	if(scene.bc.coverState == CLOSE){
	    //update cover page
	    updateCover();
	}


	scene.setCursor(cursorManager.getCursor(BACKGROUND));
	//udpate underneath pages
	updatePages(left, right);
	scene.setCursor(cursorManager.setPreviousCursor());

	//System.out.println("ThreadA.java: run() END" );
    }
    
    void updateCover()
    {
	this.frontCoverPageTexName= scene.bc.fixedTexName[BOOK_FRONT_COVER];

	textureReader= new Texture(gl,
				   glu,
				   scene.config);

		
	if( !scene.bc.front_cover_initialized ){
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: updateCover() not set before" );
	    }
	    //if not set before
	    scene.bc.front_cover_initialized= true;
	    scene.bc.front_cover_level= 1;
	    scene.bc.front_cover_imageWidth= scene.config.imageWidth_lowresol; 
	    scene.bc.front_cover_imageHeight= scene.config.imageHeight_lowresol;
	    
	    updateCoverTexture();
	    
	}
	
	else{
	    //if set before
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: updateCover() not set before" );
	    }
	    
	    if(scene.bc.front_cover_level > scene.config.maxlevel){
		System.out.println("ThreadC.java: run() reached the maximum" );
		try {
		    sleep((int)(500));
		    //suspend if required
		    synchronized(this) {
			while (threadSuspended){
			    System.out.println("ThreadC.java: run() getSuspended" );
			    wait();
			}
		    }
		}catch (InterruptedException e) { }

	    }else{
		updateCoverTexture();
	    }

	}

    }

    void updateCoverTexture()
    {
	while(scene.bc.front_cover_level <= scene.config.maxlevel && !stop){
	    try {
		sleep((int)(500));
		//suspend 
		synchronized(this) {
		    while (threadSuspended){
			if(Ttp.commentslevel >= 3){
			    System.out.println("ThreadC.java: updateCoverTexture() getSuspended" );
			}
			wait();
		    }
		}
	    } catch (InterruptedException e) { }
	    
	    if(stop) break;
	    scene.setCursor(cursorManager.getCursor(WAIT));
		
	    
	    if(!glj.gljMakeCurrent()) {
		throw new RuntimeException("ThreadC.java: putTexture(): gljMakeCurrent failed");
	    }
	    
	    textureReader.readPageTextureFromPdfReader(1, 
						       scene.bc.front_cover_imageWidth, 
						       scene.bc.front_cover_imageHeight); // rendering page 1 of pdf 
	    textureReader.setupMipmapsTexturing(frontCoverPageTexName);
	    
	    glj.gljCheckGL();
	    glj.gljFree();
	    
	    scene.bc.front_cover_imageWidth= 
		incrementImageWidth(scene.bc.front_cover_imageWidth);

	    scene.bc.front_cover_imageHeight= 
		incrementImageHeight(scene.bc.front_cover_imageHeight);
	    
	    scene.bc.front_cover_level++;
	}
    }

    void updateLeftPage(int left)
    {
	if(left < scene.config.firstPage){
	    scene.setCursor(cursorManager.setPreviousCursor());
	    return;
	}
	
	int level, imageWidth, imageHeight, texName;

	if(!texturePools.pageAlreadyinitialized(left)){
	    //level == 0
	    level= 1;
	    imageWidth= scene.config.imageWidth_lowresol;
	    imageHeight= scene.config.imageHeight_lowresol;
	    	    
	    //create a new page
	    texturePools.newPageTexture(left);
	    texName= transform(texturePools.getImageBytes(left), left);
	    
	    //update sp[1]
	    //scene.bc.sp[1].setPageUPContent(left, texName);
	    if(scene.bc.sp[1].getPage_num() < left  ){
		
	    }else{
		scene.bc.sp[1].setPageUPTexture(texName);
	    }

	    //reset back to previous Cursor state
	    //scene.setCursor(cursorManager.setPreviousCursor());

	    level++;
	    imageWidth= incrementImageWidth(imageWidth);
	    imageHeight= incrementImageHeight(imageHeight);
	    
	    //increasing resolution
	    while(level <= scene.config.maxlevel && !stop){
		try {
		    sleep((int)(500));
		    //suspend 
		    synchronized(this) {
			while (threadSuspended){
			    if(Ttp.commentslevel >= 3){
				System.out.println("ThreadA.java: updateLeftPage() getSuspended" );
			    }
			    wait();
			}
		    }
		}catch (InterruptedException e) { }
		
		if(stop) break;
		scene.setCursor(cursorManager.getCursor(WAIT));
		    
		texturePools.putTexture(left,
					level,
					imageHeight,
					imageWidth);
		texName= transform(texturePools.getImageBytes(left), left);//update the same page
		
		level++;
		imageWidth= incrementImageWidth(imageWidth);
		imageHeight= incrementImageHeight(imageHeight);
	    }//end of while
	}//end of if(!texturePools.pageAlreadyExisted(left))

	else{
	    //already exist in the cache
	    level = texturePools.getCurLevel(left);
	    imageWidth= texturePools.getCurWidth(left);
	    imageHeight= texturePools.getCurHeight(left);

	    //increment
	    level++;
	    imageWidth= incrementImageWidth(imageWidth);
	    imageHeight= incrementImageHeight(imageHeight);
	    
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: updateLeftPage() left= " + left );
		System.out.println("ThreadA.java: updateLeftPage() pick up from level= " + level );
		System.out.println("ThreadA.java: updateLeftPage() pick up from width= " + imageWidth );
		System.out.println("ThreadA.java: updateLeftPage() pick up from height= " + imageHeight );
	    }
	    
	    if(level > scene.config.maxlevel){
		//already reached maximum resolution
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: run() already reached max resolution");
		}
		
		/*try {
		    sleep((int)(500));
		    //suspend if required
		    synchronized(this) {
			while (threadSuspended){
			    System.out.println("ThreadA.java: updateLeftPage() getSuspended" );
			    wait();
			}
		    }
		    }catch (InterruptedException e) { }*/

	    }//end of if(level > scene.config.maxlevel)
	    else{
		//has not yet reached maximum resolution
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: updateLeftPage() has not yet reached max resolution");
		}
		
		while(level <= scene.config.maxlevel && !stop){
		    try {
			sleep((int)(500));
			//suspend
			synchronized(this) {
			    while (threadSuspended){
				if(Ttp.commentslevel >= 3){
				    System.out.println("ThreadA.java: updateLeftPage() getSuspended" );
				}
				wait();
			    }
			}
		    }catch (InterruptedException e) { }

		    if(stop) break;
		    scene.setCursor(cursorManager.getCursor(WAIT));

		    texturePools.putTexture(left,
					    level,
					    imageHeight,
					    imageWidth);
		    texName= transform(texturePools.getImageBytes(left), left);//update the same page
		    		    
		    level++;
		    imageWidth= incrementImageWidth(imageWidth);
		    imageHeight= incrementImageHeight(imageHeight);
		    
		}//end of while(level <= scene.config.maxlevel && !stop)
	    }//end of else
	}
    }


    /*
      Update the current left right pages
      It is not working 

    */
    void updateLeftRightPages(int left, int right)
    {
	
	int level, imageWidth, imageHeight, texName= 0;
	
	if(!texturePools.pageAlreadyinitialized(right) ||
	   !texturePools.pageAlreadyinitialized(left)){
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: both has not been initialized");
	    }

	    //does not exist in the cache
	    level= 1;
	    imageWidth= scene.config.imageWidth_lowresol;
	    imageHeight= scene.config.imageHeight_lowresol;
	    long                   starttime= 0;
	    
	    //create a new page for the left
	    if(left >= scene.config.firstPage){
		if(!texturePools.pageAlreadyinitialized(left)){
		    
		    texturePools.newPageTexture(left);
		    texName= transform(texturePools.getImageBytes(left), left);
		}
	    }
	    //create a new page for the right
	    if(right <= scene.config.lastPage){
		if(!texturePools.pageAlreadyinitialized(right)){
		    starttime = System.currentTimeMillis();
		    texturePools.newPageTexture(right);
		    texName= transform(texturePools.getImageBytes(right), right);
		    if(Ttp.commentslevel >= 3){
			System.out.println("get pages bytes Total time >  " +
					   (System.currentTimeMillis() - starttime) + " milli secs ." );
		    }
		}
	    }
	    
	    
	    //it is possible that
	    //this is called happens after bc.step_SelectedOneStaticPage(); 
	    //update sp[0]
	    //scene.bc.sp[0].setPageUPContent(right, texName);
	    if(left >= scene.config.firstPage){
		if(scene.bc.sp[1].getPage_num() < left  ){
		    
		}else{
		    scene.bc.sp[1].setPageUPTexture(texName);
		}
	    }

	    if(right <= scene.config.lastPage){
		if(scene.bc.sp[0].getPage_num() > right  ){
		    
		}else{
		    scene.bc.sp[0].setPageUPTexture(texName);
		}
	    }

	    //reset back to previous Cursor state
	    //scene.setCursor(cursorManager.setPreviousCursor());

	    level++;
	    imageWidth= incrementImageWidth(imageWidth);
	    imageHeight= incrementImageHeight(imageHeight);

	    //increasing resolution
	    while(level <= scene.config.maxlevel && !stop){
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: level= " + level );
		    System.out.println("ThreadA.java: width= " + imageWidth );
		    System.out.println("ThreadA.java: height= " + imageHeight );
		}

		try {
		    sleep((int)(500));
		    //suspend 
		    synchronized(this) {
			while (threadSuspended){
			    if(Ttp.commentslevel >= 3){
				System.out.println("ThreadA.java: updateRightPage() getSuspended" );
			    }
			    wait();
			}
		    }
		}catch (InterruptedException e) { }

		if(stop) break;
		scene.setCursor(cursorManager.getCursor(WAIT));
		
		
		if(right <= scene.config.lastPage){
		    starttime = System.currentTimeMillis();
		    texturePools.putTexture(right,
					    level,
					    imageHeight,
					    imageWidth);
		    texName= transform(texturePools.getImageBytes(right), right);//update the same page
		    if(Ttp.commentslevel >= 3){
			System.out.println("get pages bytes Total time >  " +
					   (System.currentTimeMillis() - starttime) + " milli secs ." );
		    }

		}
		
		
		if(left >= scene.config.firstPage){
		    texturePools.putTexture(left,
					    level,
					    imageHeight,
					    imageWidth);
		    texName= transform(texturePools.getImageBytes(left), left);//update the same page
		}
		
	    
		level++;
		imageWidth= incrementImageWidth(imageWidth);
		imageHeight= incrementImageHeight(imageHeight);
	    }//end of while
	}//end of if(!texturePools.pageAlreadyExisted(right))

	else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: both has been initialized");
	    }
	    
	    //already exist in the cache
	    if(right > scene.config.lastPage){
		level = texturePools.getCurLevel(left);
		imageWidth= texturePools.getCurWidth(left);
		imageHeight= texturePools.getCurHeight(left);
	    }else { 
		level = texturePools.getCurLevel(right);
		imageWidth= texturePools.getCurWidth(right);
		imageHeight= texturePools.getCurHeight(right);
	    }

	    //increment
	    level++;
	    imageWidth= incrementImageWidth(imageWidth);
	    imageHeight= incrementImageHeight(imageHeight);

	    
	    
	    if(level > scene.config.maxlevel){
		//already reached maximum resolution
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: run() already reached max resolution");
		}
		
		/*try {
		    sleep((int)(500));
		    //suspend if required
		    synchronized(this) {
			while (threadSuspended){
			    System.out.println("ThreadA.java: updateRightPage() getSuspended" );
			    wait();
			}
		    }
		    }catch (InterruptedException e) { }*/
		

	    }//end of if(level > scene.config.maxlevel)
	    else{
		//has not yet reached maximum resolution
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: updateRightPage() has not yet reached max resolution");
		}
		while(level <= scene.config.maxlevel && !stop){
		    if(Ttp.commentslevel >= 3){
			System.out.println("ThreadA.java: level= " + level );
			System.out.println("ThreadA.java: width= " + imageWidth );
			System.out.println("ThreadA.java: height= " + imageHeight );
		    }

		    try {
			sleep((int)(500));
			//suspend
			synchronized(this) {
			    while (threadSuspended){
				if(Ttp.commentslevel >= 3){
				    System.out.println("ThreadA.java: updateRightPage() getSuspended" );
				}
				wait();
			    }
			}
		    }catch (InterruptedException e) { }

		    if(stop) break;
		    scene.setCursor(cursorManager.getCursor(WAIT));

		    if(right <= scene.config.lastPage){
			if(Ttp.commentslevel >= 3){
			    System.out.println("ThreadA.java: update Right Page()" );
			}
			texturePools.putTexture(right,
						level,
						imageHeight,
						imageWidth);
			texName= transform(texturePools.getImageBytes(right), right);
		    }
		    
		    if(left >= scene.config.firstPage){
			if(Ttp.commentslevel >= 3){
			    System.out.println("ThreadA.java: update Left Page " );
			}

			texturePools.putTexture(left,
						level,
						imageHeight,
						imageWidth);
			texName= transform(texturePools.getImageBytes(left), left);//update the same page
		    }
		    
		    level++;
		    imageWidth= incrementImageWidth(imageWidth);
		    imageHeight= incrementImageHeight(imageHeight);
		    
		}//end of while(level <= scene.config.maxlevel && !stop)
		
	    }//end of else
	}

    
    }


    /*
      Update the current left page
    */
    void updateLeftPageTexture(int left)
    {
	
	int level, imageWidth, imageHeight, texName= 0;
	
	if(!texturePools.pageAlreadyinitialized(left)){
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: both has not been initialized");
	    }

	    //does not exist in the cache
	    level= 1;
	    imageWidth= scene.config.imageWidth_lowresol;
	    imageHeight= scene.config.imageHeight_lowresol;
	    long                   starttime= 0;
	    
	    //create a new page for the left
	    if(left >= scene.config.firstPage){
		if(!texturePools.pageAlreadyinitialized(left)){
		    
		    texturePools.newPageTexture(left);
		    texName= transform(texturePools.getImageBytes(left), left);
		}
	    }
	    
	    //it is possible that
	    //this is called happens after bc.step_SelectedOneStaticPage(); 
	    //update sp[0]
	    //scene.bc.sp[0].setPageUPContent(right, texName);
	    if(left >= scene.config.firstPage){
		if(scene.bc.sp[1].getPage_num() < left  ){
		    
		}else{
		    scene.bc.sp[1].setPageUPTexture(texName);
		}
	    }

	    //reset back to previous Cursor state
	    //scene.setCursor(cursorManager.setPreviousCursor());
	    level++;
	    imageWidth= incrementImageWidth(imageWidth);
	    imageHeight= incrementImageHeight(imageHeight);

	    //increasing resolution
	    while(level <= scene.config.maxlevel && !stop){
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: level= " + level );
		    System.out.println("ThreadA.java: width= " + imageWidth );
		    System.out.println("ThreadA.java: height= " + imageHeight );
		}
		
		try {
		    sleep((int)(500));
		    //suspend 
		    synchronized(this) {
			while (threadSuspended){
			    if(Ttp.commentslevel >= 3){
				System.out.println("ThreadA.java: updateRightPage() getSuspended" );
			    }

			    wait();
			}
		    }
		}catch (InterruptedException e) { }

		if(stop) break;
		scene.setCursor(cursorManager.getCursor(WAIT));
		
		
		if(left >= scene.config.firstPage){
		    texturePools.putTexture(left,
					    level,
					    imageHeight,
					    imageWidth);
		    texName= transform(texturePools.getImageBytes(left), left);//update the same page
		}
		
	    
		level++;
		imageWidth= incrementImageWidth(imageWidth);
		imageHeight= incrementImageHeight(imageHeight);
	    }//end of while
	}//end of if(!texturePools.pageAlreadyExisted(right))

	else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: both has been initialized");
	    }
	    
	    //already exist in the cache
	    
	    level = texturePools.getCurLevel(left);
	    imageWidth= texturePools.getCurWidth(left);
	    imageHeight= texturePools.getCurHeight(left);
	    

	    //increment
	    level++;
	    imageWidth= incrementImageWidth(imageWidth);
	    imageHeight= incrementImageHeight(imageHeight);

	    
	    
	    if(level > scene.config.maxlevel){
		//already reached maximum resolution
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: run() already reached max resolution");
		}
		/*try {
		    sleep((int)(500));
		    //suspend if required
		    synchronized(this) {
			while (threadSuspended){
			    System.out.println("ThreadA.java: updateRightPage() getSuspended" );
			    wait();
			}
		    }
		    }catch (InterruptedException e) { }*/
		

	    }//end of if(level > scene.config.maxlevel)
	    else{
		//has not yet reached maximum resolution
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: updateRightPage() has not yet reached max resolution");
		}
		while(level <= scene.config.maxlevel && !stop){
		    if(Ttp.commentslevel >= 3){
			System.out.println("ThreadA.java: level= " + level );
			System.out.println("ThreadA.java: width= " + imageWidth );
			System.out.println("ThreadA.java: height= " + imageHeight );
		    }

		    try {
			sleep((int)(500));
			//suspend
			synchronized(this) {
			    while (threadSuspended){
				if(Ttp.commentslevel >= 3){
				    System.out.println("ThreadA.java: updateRightPage() getSuspended" );
				}
				wait();
			    }
			}
		    }catch (InterruptedException e) { }

		    if(stop) break;
		    scene.setCursor(cursorManager.getCursor(WAIT));
		    
		    if(left >= scene.config.firstPage){
			if(Ttp.commentslevel >= 3){
			    System.out.println("ThreadA.java: update Left Page " );
			}
			
			texturePools.putTexture(left,
						level,
						imageHeight,
						imageWidth);
			texName= transform(texturePools.getImageBytes(left), left);//update the same page
		    }
		    
		    level++;
		    imageWidth= incrementImageWidth(imageWidth);
		    imageHeight= incrementImageHeight(imageHeight);
		}//end of while(level <= scene.config.maxlevel && !stop)
	    }//end of else
	}

    
    }


    /*
      Update the current right page
      It is not working 
    */
    void updateRightPageTexture(int right)
    {
	
	int level, imageWidth, imageHeight, texName= 0;
	
	if(!texturePools.pageAlreadyinitialized(right)){
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: both has not been initialized");
	    }

	    //does not exist in the cache
	    level= 1;
	    imageWidth= scene.config.imageWidth_lowresol;
	    imageHeight= scene.config.imageHeight_lowresol;
	    long                   starttime= 0;
	    
	    
	    //create a new page for the right
	    if(right <= scene.config.lastPage){
		if(!texturePools.pageAlreadyinitialized(right)){
		    texturePools.newPageTexture(right);
		    texName= transform(texturePools.getImageBytes(right), right);
		    if(Ttp.commentslevel >= 3){
			System.out.println("get pages bytes Total time >  " +
					   (System.currentTimeMillis() - starttime) + " milli secs ." );
		    }
		}
	    }
	    
	    
	    //it is possible that
	    //this is called happens after bc.step_SelectedOneStaticPage(); 
	    //update sp[0]
	    //scene.bc.sp[0].setPageUPContent(right, texName);
	    if(right <= scene.config.lastPage){
		if(scene.bc.sp[0].getPage_num() > right  ){
		    
		}else{
		    scene.bc.sp[0].setPageUPTexture(texName);
		}
	    }

	    //reset back to previous Cursor state
	    //scene.setCursor(cursorManager.setPreviousCursor());
	    level++;
	    imageWidth= incrementImageWidth(imageWidth);
	    imageHeight= incrementImageHeight(imageHeight);

	    //increasing resolution
	    while(level <= scene.config.maxlevel && !stop){
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: level= " + level );
		    System.out.println("ThreadA.java: width= " + imageWidth );
		    System.out.println("ThreadA.java: height= " + imageHeight );
		}
		try {
		    sleep((int)(500));
		    //suspend 
		    synchronized(this) {
			while (threadSuspended){
			    if(Ttp.commentslevel >= 3){
				System.out.println("ThreadA.java: updateRightPage() getSuspended" );
			    }
			    wait();
			}
		    }
		}catch (InterruptedException e) { }

		if(stop) break;
		scene.setCursor(cursorManager.getCursor(WAIT));
		
		
		if(right <= scene.config.lastPage){
		    texturePools.putTexture(right,
					    level,
					    imageHeight,
					    imageWidth);
		    texName= transform(texturePools.getImageBytes(right), right);//update the same page
		}

		level++;
		imageWidth= incrementImageWidth(imageWidth);
		imageHeight= incrementImageHeight(imageHeight);
	    }//end of while
	}//end of if(!texturePools.pageAlreadyExisted(right))

	else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: both has been initialized");
	    }
	    
	    //already exist in the cache
	    if(right > scene.config.lastPage){
		level = texturePools.getCurLevel(right-1);
		imageWidth= texturePools.getCurWidth(right-1);
		imageHeight= texturePools.getCurHeight(right-1);
	    }else { 
		level = texturePools.getCurLevel(right);
		imageWidth= texturePools.getCurWidth(right);
		imageHeight= texturePools.getCurHeight(right);
	    }

	    //increment
	    level++;
	    imageWidth= incrementImageWidth(imageWidth);
	    imageHeight= incrementImageHeight(imageHeight);

	    
	    
	    if(level > scene.config.maxlevel){
		//already reached maximum resolution
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: run() already reached max resolution");
		}
		/*try {
		    sleep((int)(500));
		    //suspend if required
		    synchronized(this) {
			while (threadSuspended){
			    System.out.println("ThreadA.java: updateRightPage() getSuspended" );
			    wait();
			}
		    }
		    }catch (InterruptedException e) { }*/
		

	    }//end of if(level > scene.config.maxlevel)
	    else{
		//has not yet reached maximum resolution
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: updateRightPage() has not yet reached max resolution");
		}
		while(level <= scene.config.maxlevel && !stop){
		    if(Ttp.commentslevel >= 3){
			System.out.println("ThreadA.java: level= " + level );
			System.out.println("ThreadA.java: width= " + imageWidth );
			System.out.println("ThreadA.java: height= " + imageHeight );
		    }

		    try {
			sleep((int)(500));
			//suspend
			synchronized(this) {
			    while (threadSuspended){
				if(Ttp.commentslevel >= 3){
				    System.out.println("ThreadA.java: updateRightPage() getSuspended" );
				}
				wait();
			    }
			}
		    }catch (InterruptedException e) { }

		    if(stop) break;
		    scene.setCursor(cursorManager.getCursor(WAIT));

		    if(right <= scene.config.lastPage){
			if(Ttp.commentslevel >= 3){
			    System.out.println("ThreadA.java: update Right Page()" );
			}
			texturePools.putTexture(right,
						level,
						imageHeight,
						imageWidth);
			texName= transform(texturePools.getImageBytes(right), right);
		    }

		    level++;
		    imageWidth= incrementImageWidth(imageWidth);
		    imageHeight= incrementImageHeight(imageHeight);
		    
		}//end of while(level <= scene.config.maxlevel && !stop)
		
	    }//end of else
	}

    
    }

    
    void updatePages(int left, int right)
    {
	
	//starting page
	/*if(scene.bc.coverState == OPEN){
	    if(right == scene.config.firstPage){
		right++;
		//left unchanged 
	    }
	    
	    else if(left == scene.config.lastPage){
		left--;
	    }

	    else{
		right++;
		left--;
	    }
	}
	else if(scene.bc.coverState == CLOSE){
	    if(scene.bc.isFrontCoverFacingUP()){
		//right == 0  unchanged
		//left == -1 unchanged
	    }else{
		//left == lastPage unchanged
		//right == lastPage+1  unchanged
	    }
	    }*/
	
	while( (right <= scene.config.lastPage || left >= scene.config.firstPage) && !stop ){
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadB.java: run() running" );
	    }
	    
	    //updateLeftRightPages(left, right);

	    updateLeftPageTexture(left);
	    updateRightPageTexture(right);
	    	    
	    /*try {
		sleep((int)(500));
		
		synchronized(this) {
		    while (threadSuspended){
			System.out.println("ThreadB.java: run() getSuspended" );
                        wait();
		    }
                }
	    } catch (InterruptedException e) {
		System.out.println("ThreadB.java: run() Interrupt" );  
	    }
	    
	    if(stop) break;
	    scene.setCursor(cursorManager.getCursor(BACKGROUND));
	    */
	    //right
	    if(right <= scene.config.lastPage){
		//texturePools.newPageTexture(right);
		right++;
	    }
	    
	    //left
	    if(left >= scene.config.firstPage){
		//texturePools.newPageTexture(left);
		left--;
	    }
	    
	}


    }


    int transform(byte [] imagebytes, int page_num)
    {
	int result;
	
	if(imagebytes == null){
	    result= 0;
	    return result;
	}

	
	if(!glj.gljMakeCurrent()) {
	    throw new RuntimeException("ThreadA.java: putTexture(): gljMakeCurrent failed");
	}

	synchronized(texturePools){
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: transform(): synchronized START" );
	    }
	    
	    result= texturePools.getTexName(page_num);
	
	    if(result == 0){
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: transform(): generate new texture" );
		}
		//generate new texture 
		int [] texNames= new int[1];
		gl.glGenTextures(1, texNames);
		result= texNames[0];// new texture name
		//update cache
		texturePools.setTexName(page_num, result);
		
	    }else{
		if(Ttp.commentslevel >= 3){
		    System.out.println("ThreadA.java: transform(): update the existing texture= " + result);
		}
		//update the existing texture
	    }
	    
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: transform(): synchronized END" );
	    }
	}
	
	//load image bytes from cache to GLContext
	if(textureReader.readPageTextureFromInputStream(new ByteArrayInputStream(imagebytes, 0, imagebytes.length))){
	    textureReader.setupMipmapsTexturing(result);
	}else{
	    if(Ttp.commentslevel >= 3){
		System.out.println("ThreadA.java: transform(): Fail to read from image buffer" );
	    }
	    result= 0;
	}
	
	glj.gljCheckGL();
	glj.gljFree();
	
	if(Ttp.commentslevel >= 3){
	    System.out.println("ThreadA.java: transform(): Successfully update texture" );
	}


	return result;
    }

    public void setStop()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("\tThreadA.java: setStop()");
	}
	this.stop= true;
	//interrupt();
    }
    
    public void setSuspendflipflop()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("\tThreadA.java: setSuspendflipflop()");
	}
	threadSuspended = !threadSuspended;

        if (!threadSuspended){
	    synchronized(this) {
		notifyAll();
	    }
	}
    }

    public synchronized void setSuspend()
    {
	if(Ttp.commentslevel >= 2){
	    System.out.println("\tThreadA.java: setSuspend()");
	}
	this.threadSuspended = true;
    }

    public synchronized void setAwake()
    {
	if(Ttp.commentslevel >= 3){
	    System.out.println("\tThreadA.java: setAwake()");
	}
	this.threadSuspended = false;
	notifyAll();
    }
   
    public synchronized boolean isSuspended()
    {
	return threadSuspended;
    }

 
    int incrementImageHeight(int height)
    {
	int newHeight;
	
	if(scene.config.increMethod == ADD){ 
	    //through addition
	    newHeight = height + scene.config.imageDelta;
	}else{
	    //through multiplication
	    newHeight= height * scene.config.imageMultiple;
	}
		
	return newHeight;
    }
    
    int incrementImageWidth(int width)
    {
	int newWidth;
	
	if(scene.config.increMethod == ADD){ 
	    //through addition
	    newWidth = width + scene.config.imageDelta;
	}else{
	    //through multiplication
	    newWidth= width * scene.config.imageMultiple;
	}
	
	
	return newWidth;
    }
}
